""" Bid for Game Hand module."""
import os
import json
import inspect
from .bidding import Bid, Pass
from bridgeobjects import Hand, Card, Call, Suit, NoTrumps, SUIT_NAMES, CALLS, SUITS


class Hand(Hand):
    """A sub class of bridgeobjects Hand, to enable bidding."""
    MODULE_TRACE = 'batch_tests/data//module_trace.txt'
    OVERCALLER_POSITION = {'none': 0, 'second_seat': 1, 'fourth_seat': 2}

    spade_suit, heart_suit, diamond_suit, club_suit = SUITS['S'], SUITS['H'], SUITS['D'], SUITS['C']

    def __init__(self, hand_cards=None, board=None, *args, **kwargs):
        super().__init__(hand_cards, *args, **kwargs)
        self.board = board
        self.suits = [SUITS[name] for name in 'CDHS']
        self.no_trumps = NoTrumps()
        if board:
            self.bid_history = board.active_bid_history
            if len(self.bid_history) % 2 == 0:
                self.overcaller = False
            else:
                self.overcaller = True
        else:
            self.bid_history = []
            self.overcaller = False
        self._losers = None
        self.overcaller_in_second_seat = False
        self.overcaller_in_fourth_seat = False

        # Initialise specific bids
        pass_bid = Pass()
        self.opener_bid_one = pass_bid
        self.opener_bid_two = pass_bid
        self.opener_bid_three = pass_bid
        self.responder_bid_one = pass_bid
        self.responder_bid_two = pass_bid
        self.responder_bid_three = pass_bid
        self.overcaller_bid_one = pass_bid
        self.overcaller_bid_two = pass_bid
        self.overcallers_last_bid = pass_bid
        self.advancer_bid_one = pass_bid
        self.advancer_bid_two = pass_bid
        self.previous_bid = pass_bid
        self.last_bid = pass_bid
        self.right_hand_bid = pass_bid
        self.my_last_bid = pass_bid
        self.partner_bid_one = pass_bid
        self.partner_bid_two = pass_bid
        self.partner_last_bid = pass_bid
        self.bid_one = pass_bid
        self.bid_two = pass_bid
        self.bid_three = pass_bid
        self.unplayed_cards = [card for card in self.cards]

        self.responders_support = False
        self.overcaller_has_jumped = False
        self.bid_after_stayman = False

        self.assign_bids()
        self._update_longest_suit()
        self.trace_module, self.last_hand = self._get_trace_module()

    def to_json(self):
        """Return object as json string property."""
        json_str = json.dumps({
            'cards': [card.name for card in self.cards],
            'unplayed_cards': [card.name for card in self.unplayed_cards],
        })
        return json_str

    def from_json(self, json_str):
        """Populate the attributes from the json string."""
        hand_dict = json.loads(json_str)
        self.cards = [Card(name) for name in hand_dict['cards']]
        self.unplayed_cards = [Card(name) for name in hand_dict['unplayed_cards']]

    def _update_longest_suit(self):
        """Change the longest suit if partner has bid it."""
        if 4 <= self.shape[1] == self.shape[0]:
            if 1 < len(self.bid_history) <= 4:
                partners_suit = Bid(self.bid_history[-2]).denomination
                if self.longest_suit == partners_suit and self.second_suit not in self. opponents_suits:
                    self._longest_suit, self._second_suit = self.second_suit, self.longest_suit

    @property
    def longest_suit(self):
        """Return the longest_suit."""
        return self._longest_suit

    @property
    def second_suit(self):
        """Return the second suit."""
        return self._second_suit

    @property
    def nt_level(self):
        """Return the level of no trumps"""
        return self.next_nt_bid().level

    @property
    def losers(self):
        """Return the number of losers in the Hand."""
        if not self._losers:
            self._losers = self._get_losers()
        return self._losers

    def _get_losers(self):
        """Calculate and return the number of losers in the Hand."""
        losers = 0
        losers_by_suit = {
            'S': [0, 0],
            'H': [0, 0],
            'D': [0, 0],
            'C': [0, 0],
        }
        for suit in SUIT_NAMES:
            cards = [card for card in self.cards_by_suit[suit]]
            honour_count = 0
            for card in cards:
                if card.value >= 12:
                    honour_count += 1
                elif card.value == 11:
                    if self.suit_points(suit) > 2:
                        honour_count += 1
            count = min(3, len(cards))
            losers += count - honour_count
        return losers

    def assign_bids(self):
        """Assign bids based on bid_history."""
        bids = len(self.bid_history)

        if bids >= 1:
            self.opener_bid_one = Bid(self.bid_history[0])
            self.last_bid = Bid(self.bid_history[-1])
            if bids == 1:
                return

        if bids >= 2:
            self._assign_overcaller_seat()
            self.overcaller_bid_one = Bid(self.bid_history[1])
            # noinspection PyTypeChecker
            self.overcaller_has_jumped = self.is_jump(self.opener_bid_one,
                                                      self.overcaller_bid_one)
            self.right_hand_bid = Bid(self.bid_history[-1])
            self.partner_last_bid = Bid(self.bid_history[-2])
            self.partner_bid_one = Bid(self.bid_history[-2])
            if bids == 2:
                return

        if bids >= 3:
            self.responder_bid_one = Bid(self.bid_history[2])
            self.responders_support = (self.opener_bid_one.denomination ==
                                       self.responder_bid_one.denomination)
            if bids == 3:
                return

        if bids >= 4:
            self.advancer_bid_one = Bid(self.bid_history[3])
            self.bid_one = Bid(self.bid_history[-4])
            self.previous_bid = Bid(self.bid_history[-4])
            if self.overcaller_in_fourth_seat:
                self.overcaller_bid_one = Bid(self.bid_history[3])
                # noinspection PyTypeChecker
                self.overcaller_has_jumped = self.is_jump(self.responder_bid_one,
                                                          self.overcaller_bid_one)
            self.my_last_bid = Bid(self.bid_history[-4])
            if bids == 4:
                return

        if bids >= 5:
            self.opener_bid_two = Bid(self.bid_history[4])
            self.bid_after_stayman = self._bid_after_stayman()
            if bids == 5:
                return

        if bids >= 6:
            self.partner_bid_one = Bid(self.bid_history[-6])
            self.partner_bid_two = Bid(self.bid_history[-2])
            if self.overcaller_in_fourth_seat:
                self.overcaller_bid_one = Bid(self.bid_history[3])
                self.advancer_bid_one = Bid(self.bid_history[5])
            if bids == 6:
                return

        if bids >= 7:
            self.responder_bid_two = Bid(self.bid_history[6])
            self.overcaller_bid_two = Bid(self.bid_history[5])
            if bids == 7:
                return

        if bids >= 8:
            self.bid_one = Bid(self.bid_history[-8])
            self.bid_two = Bid(self.bid_history[-4])
            self.overcallers_last_bid = Bid(self.bid_history[-1])
            if self.overcaller_in_fourth_seat:
                self.overcaller_bid_two = Bid(self.bid_history[7])
            else:
                self.advancer_bid_two = Bid(self.bid_history[7])
            if bids == 8:
                return

        if bids >= 9:
            self.opener_bid_three = Bid(self.bid_history[8], '')
            if bids == 9:
                return

        if bids >= 10:
            self.partner_bid_one = Bid(self.bid_history[-10])
            self.partner_bid_two = Bid(self.bid_history[-6])

        if bids >= 11:
            self.responder_bid_three = Bid(self.bid_history[10])
            # self.overcaller_bid_two = Bid(self.bid_history[5])

        if bids >= 12:
            self.bid_one = Bid(self.bid_history[-12])
            self.bid_two = Bid(self.bid_history[-8])
            self.bid_three = Bid(self.bid_history[-4])

    def _assign_overcaller_seat(self):
        """Assign seat to overcaller."""
        if len(self.bid_history) >= 2:
            if Bid(self.bid_history[1]).name != Pass().name:
                self.overcaller_in_second_seat = True
            elif (len(self.bid_history) >= 4 and
                    not Bid(self.bid_history[3]).is_pass):
                self.overcaller_in_fourth_seat = True

    def _get_trace_module(self):
        """Return the name of the module to trace."""
        if os.path.isfile(self.MODULE_TRACE):
            with open(self.MODULE_TRACE, 'r') as f_trace_file:
                text = f_trace_file.read()
                text = text.split('\n')
                text.extend(['', ''])
                return text[0], text[1]
        else:
            return None, None

    def _set_trace_hand(self):
        """Write the current hand to the trace file."""
        if os.path.isfile(self.MODULE_TRACE):
            with open(self.MODULE_TRACE, 'w') as f_trace_file:
                f_trace_file.write('\n'.join([self.trace_module, self.__str__()]))

    def is_insufficient_bid(self, test_bid):
        """Return True if the bid is insufficient."""
        last_bid = None
        for last_bid in self.bid_history[::-1]:
            if Bid(last_bid).is_value_call:
                break
        last_bid_index = CALLS.index(last_bid)
        test_bid_index = CALLS.index(test_bid.name)
        if test_bid_index <= last_bid_index:
            result = True
        else:
            result = False
        return result

    @property
    def singleton_honour(self):
        """Return True if hand has singleton K or Q."""
        return self._singleton_honour()

    def _singleton_honour(self):
        """Return True if hand has singleton K or Q."""
        if self.shape[3] == 1:
            for suit_name in SUIT_NAMES:
                suit = Suit(suit_name)
                if self.suit_length(suit) == 1:
                    for card in self.cards:
                        if card.suit == suit:
                            if card.rank in 'KQ':
                                return True
        return False

    @property
    def opponents_have_bid(self):
        """Return True if the opponents have made a bid."""
        return self._opponents_have_bid()

    def _opponents_have_bid(self):
        """Return True if the opponents have made a bid."""
        for bid in self.bid_history[1::2]:
            if Bid(bid).is_value_call:
                return True
        return False

    @property
    def competitive_auction(self):
        """Return True if the last bid is not Pass or Double."""
        return self._competitive_auction()

    def _competitive_auction(self):
        """Return True if the last bid is not Pass or Double."""
        value = False
        if len(self.bid_history) >= 1:
            value = Bid(self.bid_history[-1]).is_value_call
        if len(self.bid_history) >= 3:
            value = value or Bid(self.bid_history[-3]).is_value_call
        if len(self.bid_history) >= 5:
            value = value or Bid(self.bid_history[-5]).is_value_call
        return value

    @staticmethod
    def _is_value_call_or_double(bid):
        """Return True if is value call or double."""
        call = Bid(bid)
        value = call.is_value_call or call.is_double
        return value

    @property
    def opponents_have_doubled(self):
        """Return True if opponents have doubled with no further bid from them."""
        return self._opponents_have_doubled()

    def _opponents_have_doubled(self):
        """Return True if opponents have doubled with no further bid from them."""
        value = False
        if len(self.bid_history) >= 1:
            if Bid(self.bid_history[-1]).is_double:
                value = True
            elif len(self.bid_history) >= 3:
                if(Bid(self.bid_history[-3]).is_double and
                        Bid(self.bid_history[-1]).is_pass):
                    value = True
        return value

    @property
    def partner_has_passed(self):
        """Return True of partner passed on first round."""
        return self._partner_has_passed()

    def _partner_has_passed(self):
        """Return True of partner passed on first round."""
        value = False
        if len(self.bid_history) >= 5:
            if Bid(self.bid_history[-5]).is_pass:
                value = True
        return value

    def _bid_after_stayman(self):
        """Return True if responder has bid Clubs after NT opening."""
        result = False
        if (self.opener_bid_one.is_nt and
                self.responder_bid_one.denomination == self.club_suit):
            result = True
        return result

# TODO: sort out use of shortage points
    def suit_bid(self, level, suit, comment='0000', use_shortage_points=False):
        """Return a bid in the suit at the given level."""
        return Bid(self._call_name(level, suit.name), comment, use_shortage_points)

    def heart_bid(self, level, comment='0000', use_shortage_points=False):
        """Return a bid in Hearts at the given level."""
        return Bid(self._call_name(level, 'H'), comment, use_shortage_points)

    def spade_bid(self, level, comment='0000', use_shortage_points=False):
        """Return a bid in Spades at the given level."""
        return Bid(self._call_name(level, 'S'), comment, use_shortage_points)

    def club_bid(self, level, comment='0000', use_shortage_points=False):
        """Return a bid in Clubs at the given level."""
        return Bid(self._call_name(level, 'C'), comment, use_shortage_points)

    def diamond_bid(self, level, comment='0000', use_shortage_points=False):
        """Return a bid in Diamonds at the given level."""
        return Bid(self._call_name(level, 'D'), comment, use_shortage_points)

    def nt_bid(self, level, comment='0000', use_shortage_points=False):
        """Return a bid in NT at the given level."""
        return Bid(self._call_name(level, 'NT'), comment, use_shortage_points)

    def barrier_is_broken(self, first_bid, second_bid):
        """Return True if second_bid breaks the barrier relative to bid_one."""
        broken = False
        level_one = first_bid.level
        barrier = Bid(self._call_name(level_one+1, first_bid.denomination.name))
        # noinspection PyTypeChecker
        if self._higher_bid(barrier, second_bid):
            broken = True
        return broken

    def cheapest_long_suit(self):
        """Return the longest suit or cheapest of equal length suits to bid."""
        if self.shape[0] > self.shape[1] and self.longest_suit not in self.opponents_suits:
            return self.longest_suit
        else:
            max_length = self.shape[1]
            last_bid_denomination = self._last_denomination_called()
            extended_suit_list = self._get_extended_suit_list()

            # find the starting index in the extended suit list
            index = extended_suit_list.index(last_bid_denomination)

            # find the level of NT in the middle of the extended suit list
            nt_level = self.nt_level
            while index < len(extended_suit_list) - 1 and index < 8:
                index += 1
                if index >= 4:
                    # if we're in the second half, increment the level by 1.
                    # I.e. we have past NT and onto the next level
                    level =  nt_level + 1
                else:
                    level = nt_level

                # test the next suit
                suit = extended_suit_list[index]
                if (self.suit_length(suit) == max_length and
                        self.next_level(suit) == level and
                        suit not in self.opponents_suits):
                    return suit
            else:
                return None

    def _last_denomination_called(self):
        """Return the denomination of the last value call."""
        for bid in self.bid_history[::-1]:
            call = Bid(bid)
            if call.is_suit_call:
                return call.denomination
        else:
            assert False, 'Last denomination called with no value call in history'

    @staticmethod
    def _get_extended_suit_list():
        """Return an extended list of sorted of suits, e.g. [C, D, H, S, C, D, H, S]."""
        extended_suit_list = [Suit(name) for name in SUIT_NAMES]
        extended_suit_list.sort(key=lambda x: x.rank)

        # get a double suit list, e.g. [C, D, H, S, C, D, H, S]
        extended_suit_list.extend(extended_suit_list)
        return extended_suit_list

    @staticmethod
    def _higher_bid(first_bid, second_bid):
        """Return True if first_bid is the lower."""
        higher = False
        level_one = first_bid.level
        level_two = second_bid.level
        if level_two > level_one:
            higher = True
        else:
            if level_one == level_two:
                if first_bid.denomination < second_bid.denomination:
                    higher = True
        return higher

    @staticmethod
    def is_jump(first_bid, second_bid):
        """Return True if second bid is a jump over first."""
        # e.g.  first_bid = Bid('1S')
        #       second_bid = Bid('2NT')
        result = False
        if second_bid.is_value_call and first_bid.is_value_call:
            jump_level = second_bid.level - first_bid.level
            if jump_level > 1:
                result = True
            elif jump_level == 1:
                if second_bid.is_suit_call and first_bid.is_suit_call:
                    if second_bid.denomination > first_bid.denomination:
                        result = True
                elif second_bid.is_nt:
                    result = True
            else:
                result = False
        return result

    def jump_bid_made(self, test_bid):
        """Test bid against bid-history and return True
           if test_bid is a jump over last relevant bid.
        """
        second_bid = None
        for bid in self.bid_history[::-1]:
            if bid != 'P' and bid != 'D' and bid != 'R':
                second_bid = Bid(bid)
                break
        jump_level = test_bid.level - second_bid.level
        if jump_level > 1:
            result = True
        elif jump_level == 1 and second_bid.denomination < test_bid.denomination:
            result = True
        else:
            result = False
        return result

    def levels_to_bid(self, bid_one, bid_two):
        """
        Return the number of 'steps' between bid_one and bid_two.

        e.g.

        1C to 3H is 3 steps: (1C, 2C 3C, 3H) but
        1S to 3H is 2 steps: (1S, 2S, 3H).
        """
        steps = 0
        while bid_one < bid_two:
            bid_one = self.suit_bid(bid_one.level+1, bid_one.denomination)
            steps += 1
        return steps

    @property
    def overcall_made(self):
        """Return 1 if an overcall has been made by 2nd and not by 4th seat.
        Return 2 if overcall has been made in 4th seat.
        """
        return self._overcall_made()

    def _overcall_made(self):
        """Return 1 if an overcall has been made by 2nd and not by 4th seat.
        Return 2 if overcall has been made in 4th seat.
        """
        bid_history = self.bid_history
        result = self.OVERCALLER_POSITION['none']
        for bid in bid_history[1::2]:
            if not Bid(bid).is_pass:
                result = self.OVERCALLER_POSITION['second_seat']
                if Bid(bid_history[-1]).is_value_call and len(bid_history) > 2:
                    result = self.OVERCALLER_POSITION['fourth_seat']
                break
        return result

    @property
    def opponents_at_game(self):
        """Return True if opponents at game level."""
        return self._opponents_at_game()

    def _opponents_at_game(self):
        """Return True if opponents at game level."""
        result = False
        if len(self.bid_history) >= 3:
            if (Bid(self.bid_history[-3]).is_game or
                    Bid(self.bid_history[-1]).is_game):
                result = True
        elif len(self.bid_history) >= 1:
            if Bid(self.bid_history[-1]).is_game:
                result = True
        return result

    @property
    def bidding_above_game(self):
        """Return True if the bidding is at or above game level."""
        return self._bidding_above_game()

    def _bidding_above_game(self):
        """Return True if the bidding is at or above game level."""
        result = False
        for bid in self.bid_history[::1]:
            if Bid(bid).is_game:
                result = True
                break
        return result

    @property
    def partner_doubled_game(self):
        """Return True if partner has doubled at or above game level."""
        return self._partner_doubled_game()

    def _partner_doubled_game(self):
        """Return True if partner has doubled at or above game level."""
        result = False
        if 'D' in self.bid_history:
            index = self.bid_history.index('D')
            bids = [Bid(bid) for bid in self.bid_history[index::-1]]
            for bid in bids:
                if bid.is_value_call:
                    result = bid.is_game
                    break
        return result

    def double_level(self):
        """Return the level at which the DOUBLE was made."""
        level = 0
        index = self.bid_history.index('D')
        bids = [Bid(bid) for bid in self.bid_history[index::-1]]
        for bid in bids:
            if bid.is_value_call:
                level = bid.level
                break
        return level

    @property
    def opponents_suits(self):
        """Return a list of all opponent's bid suits."""
        return self._opponents_suits()

    def _opponents_suits(self):
        """Return a list of all opponent's bid suits."""
        opponents_suits = []
        if self.overcaller:
            start = 0
        else:
            start = 1
        for bid in self.bid_history[start::2]:
            call = Bid(bid)
            if call.is_suit_call:
                opponents_suits.append(Suit(call.denomination.name))
        return opponents_suits

    def partner_bids_at_lowest_level(self):
        """Return True if partner has not jumped."""
        value = False
        double_index = None
        for index, bid in enumerate(self.bid_history):
            if Bid(bid).is_double:
                double_index = index
                break
        partners_bid = Bid(self.bid_history[double_index+2])
        for bid in list(reversed(self.bid_history))[double_index+1::2]:
            if Bid(bid).is_value_call:
                # noinspection PyTypeChecker
                value = self.is_jump(Bid(bid), partners_bid)
        return value

    def openers_agreed_suit(self):
        """Return a suit agreed by opener."""
        suit = None
        if len(self.bid_history) >= 5:
            opponents_bid_one = Bid(self.bid_history[0], '')
            opponents_bid_two = Bid(self.bid_history[2], '')
            opponents_bid_three = Bid(self.bid_history[4], '')
            if ((opponents_bid_one.denomination == opponents_bid_two.denomination) or
                    (opponents_bid_two.denomination == opponents_bid_three.denomination)):
                suit = opponents_bid_two.denomination
            if opponents_bid_one.is_double and opponents_bid_two.is_suit_call:
                suit = opponents_bid_two.denomination
        return suit

    def cheaper_suit(self, suit_one, suit_two):
        """Return the cheaper of two suits based on bid history."""
        suits = [suit_one, suit_two]
        suit = self.cheapest_suit(suits)
        return suit

    def next_four_card_suit(self):
        """Return the cheapest four card suit."""
        suits = []
        for suit in self.suits_by_length:
            if self.suit_length(suit) == 4:
                suits.append(suit)
        suit = self.cheapest_suit(suits)
        return suit

    def cheapest_suit(self, suits):
        """Return the cheapest suit based on bid history."""
        suit_names = [(suit.name, suit.rank) for suit in suits]
        sorted_suits = sorted(suit_names, key=lambda tup: tup[1])
        suits = [Suit(suit[0]) for suit in sorted_suits]
        last_rank = self._last_bid_rank()
        if not suits:
            return None
        for suit in suits:
            if suit.rank > last_rank:
                break
        else:
            suit = suits[0]
        return suit

    def _last_bid_rank(self):
        """Return the rank of the last value bid."""
        bid = None
        for bid in self.bid_history[::-1]:
            if Call(bid).is_value_call:
                break
        last_rank = Call(bid).denomination.rank
        if last_rank == 4:
            last_rank = -1
        return last_rank

    @staticmethod
    def game_level(suit):
        """Return the level of game in the suit."""
        if suit.is_major:
            level = 4
        elif suit.is_minor:
            level = 5
        else:
            level = 3
        return level

    def bid_to_game(self, denomination, comment='0000', use_distribution_points=False):
        """Return game level bid in given denomination"""
        if denomination.is_nt:
            bid = Bid('3NT', comment, use_distribution_points)
        elif denomination.is_major:
            bid = Bid(self._call_name(4,  denomination.name), comment, use_distribution_points)
        elif denomination.is_minor:
            bid = Bid(self._call_name(5,  denomination.name), comment, use_distribution_points)
        else:
            bid = None
        return bid

    @property
    def stoppers_in_bid_suits(self):
        """Return True if hand contains stoppers in all opponent's bid suits."""
        return self._stoppers_in_bid_suits()

    @property
    def poor_stoppers_in_bid_suits(self):
        """Return True if hand contains stoppers (including ten) in all opponent's bid suits."""
        return self._stoppers_in_bid_suits(lowest_card='T')

    def _stoppers_in_bid_suits(self, lowest_card='J'):
        """Return True if hand contains stoppers in all opponent's bid suits."""
        result = True
        for suit in self.opponents_suits:
            if not self.suit_stopper(suit, lowest_card):
                result = False
                break
            if not result:
                break
        return result

    def suit_stopper(self, suit, lowest_card='J'):
        """Return True if the hand contains a stopper in 'suit'."""
        suit_holding = self.suit_holding
        result = False
        if not suit.is_suit:
            return False
        if suit_holding[suit] >= 5:
            result = True
        else:
            poor_stopper = False
            ace_stopper = Card('A', suit.name) in self.cards
            king_stopper = (Card('K', suit.name) in self.cards and suit_holding[suit] >= 2)
            queen_stopper = (Card('Q', suit.name) in self.cards and suit_holding[suit] >= 3)
            jack_stopper = (Card('J', suit.name) in self.cards and suit_holding[suit] >= 4)
            if lowest_card == 'T':
                poor_stopper = (Card('T', suit.name) in self.cards and suit_holding[suit] >= 4)
            if ace_stopper or king_stopper or queen_stopper or jack_stopper or poor_stopper:
                result = True
        return result

    @property
    def unbid_suit(self):
        """Return the unbid _suit (if any) or None."""
        return self._unbid_suit()

    def _unbid_suit(self):
        """Return the unbid _suit (if any) or None."""
        suits = [suit for suit in self.suits]
        for bid in self.bid_history:
            bid_suit = Bid(bid).denomination
            if bid_suit in suits:
                suits.remove(bid_suit)
        if len(suits) == 1:
            return suits[0]
        else:
            return None

    def stoppers_in_unbid_suits(self):
        """Return True if hand contains stoppers in all opponent's unbid suits."""
        result = True
        bid_suits = []
        for bid in self.bid_history:
            bid_suit = Bid(bid).denomination
            if bid_suit.is_suit:
                bid_suits.append(bid_suit)
        for suit in self.suits:
            if suit not in bid_suits:
                if not self.suit_stopper(suit):
                    result = False
                    break
            if not result:
                break
        return result

    def stoppers_in_other_suits(self, suit):
        """Return True if hand has stoppers in all suits except suit."""
        stoppers = True
        for test_suit in self.suits:
            if test_suit != suit:
                if not self.suit_stopper(test_suit):
                    stoppers = False
                    break
        return stoppers

    def four_in_bid_suits(self, lowest_card='J'):
        """Return True if hand contains stoppers or 4 cards in all
            opponent's bid suits.
        """
        result = True
        for suit in self.opponents_suits:
            if (not self.suit_stopper(suit, lowest_card) and
                    self.suit_holding[suit] <= 3):
                result = False
                break
            if not result:
                break
        return result

    def five_in_opponents_suits(self):
        """Return True if hand has five cards in opponents suits."""
        result = False
        for suit in self.opponents_suits:
            if self.suit_holding[suit] >= 5:
                result = True
        return result

    def three_suits_bid_and_stopper(self):
        """Returns True if three suits bid and
                hand has an stopper in the unbid suit"""
        result = False
        suits_bid = [False, False, False, False]
        for bid_name in self.bid_history[::2]:
            bid = Bid(bid_name)
            if bid.is_suit_call:
                suits_bid[bid.denomination.rank] = True
                suits_bid[bid.denomination.rank] = True
        if suits_bid.count(True) == 3:
            for index, suit_bid in enumerate(suits_bid):
                if not suit_bid:
                    suit = self.suits[index]
                    if self.suit_stopper(suit):
                        result = True
                    elif self.suit_points(suit) >= 1 and self.suit_holding[suit] >= 3:
                        result = True
                    break
        return result

    def can_bid_suit(self, suit):
        """Return False if suit is in opponents bids."""
        result = True
        for bid in self.bid_history[::-1][::2]:
            if Bid(bid).is_suit_call:
                if suit == Bid(bid).denomination:
                    result = False
                    break
        return result

    def has_stopper(self, suit):
        """Return self.suit_stopper."""
        return self.suit_stopper(suit)

    def next_level(self, suit, raise_level=0):
        """Return the next next level for a suit bid."""
        level = self.next_level_bid(suit, '000', raise_level).level
        return level

    def current_bid_level(self):
        """Return the level of the latest quantitative bid."""
        bid = self._get_last_bid()
        return bid.level

    def next_level_bid(self, suit, comment='0000', raise_level=0):
        """Return the lowest possible bid in suit."""
        last_bid = self._get_last_bid()
        if last_bid is None:
            level = 1
        else:
            level = last_bid.level
            level += raise_level
            if last_bid.is_nt:
                level += 1
            elif suit.is_suit:
                if last_bid.denomination >= suit:
                    level += 1
            if level > 7:
                level = 7
        new_bid = Bid(self._call_name(level, suit.name), comment)
        return new_bid

    def _get_last_bid(self):
        """Return last quantitative bid from history."""
        last_bid = None
        for bid_level in self.bid_history[::-1]:
            if bid_level != 'P' and bid_level != 'D' and bid_level != 'R':
                last_bid = Bid(bid_level, '0000')
                break
                break
        return last_bid

    def next_nt_bid(self, comment='0000', raise_level=0):
        """Return the lowest possible bid in no trumps."""
        denomination = self.no_trumps
        bid = self.next_level_bid(denomination, comment, raise_level)
        return bid

    def responder_weak_bid(self):
        """Responder has shown preference at  the lowest level."""
        overcallers_last_bid = Call(self.bid_history[-3])
        if (overcallers_last_bid.is_pass and
                self.partner_bid_one.is_pass and
                (self.partner_last_bid.denomination == self.opener_bid_one.denomination or
                 self.partner_last_bid.denomination == self.opener_bid_two.denomination) and
                not self.is_jump(self.opener_bid_two, self.partner_last_bid)):
            return True
        else:
            return False

    def advancer_preference(self, call_id='0000'):
        """Respond after a 3 level bid make suit preference."""
        if self.overcaller_bid_one.is_double:
            suit_one = self.overcaller_bid_two.denomination
            suit_two = self.overcaller_bid_three.denomination
        else:
            suit_one = self.overcaller_bid_one.denomination
            suit_two = self.overcaller_bid_two.denomination
        if suit_one.is_major and self.suit_length(suit_one) >= 3:
            suit = suit_one
        elif self.suit_holding[suit_one] + 1 >= self.suit_holding[suit_two]:
            suit = suit_one
        else:
            suit = suit_two
        raise_level = 0
        if self.hcp >= 8 and self.next_level(suit) <= 3:
            raise_level = 1
        bid = self.next_level_bid(suit, call_id, raise_level=raise_level)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def unbid_four_card_major(self):
        """Return an unbid four card major or None."""
        suit = None
        bid_suits = [Bid(bid).denomination for bid in self.bid_history if Bid(bid).is_suit_call]
        if (self.hearts >= 4 and
                self.heart_suit not in bid_suits):
            suit = self.heart_suit
        elif (self.spades >= 4 and
                self.spade_suit not in bid_suits):
            suit = self.spade_suit
        return suit

    @staticmethod
    def quantitative_raise(points, base_level,
                           point_list, maximum_level=5):
        """
            Return a bid based on a quantitative raise.
            max raise is the number of elements in pointlist 1, 2,3 or 4
            scan the (reversed) points list until the points in
            the hand exceeds the level
            This shows whether it is a 3,2 or 1 raise etc.

            e.g.
            level = self.quantitative_raise(points, 1, [6, 10, 13, 16], 5)
            if points = 11 this raises level = 1+2 = 3.
        """
        maximum_raise = len(point_list)
        point_list = list(reversed(point_list))
        raise_level = 0
        for index, item in enumerate(point_list):
            if points >= item:
                raise_level = base_level + maximum_raise - index
                break
        if raise_level > maximum_level:
            raise_level = maximum_level
        return raise_level

    def hand_value_points(self, bid_suit):
        """Return the hand value points for the given suit."""
        hand_value_points = (self.hcp + self.support_shape_points(bid_suit))
        return hand_value_points

    def support_points(self, bidders_suit):
        """
        Calculate the sum of high card points and distribution points
        based on support_shape_points.
        """
        return self.high_card_points + self.support_shape_points(bidders_suit)

    def support_shape_points(self, bidders_suit):
        """
        Calculate the distribution points based on
        3 for a void, 2 for a singleton and 1 for a doubleton.
        """
        points = 0
        if bidders_suit.is_suit:
            if self._suit_support(bidders_suit):
                for index in range(4):
                    if self.shape[index] < 3:
                        points += 3 - self.shape[index]
        return points

    def _suit_support(self, bid_suit):
        """Return True if the hand contains at least 3 of bid_suit."""
        match = False
        if self.suit_holding[bid_suit] >= 3:
            match = True
        return match

    @property
    def ordered_holding(self):
        """Returns suits and holdings in decreasing order of holding
            e.g. [[5, 1], [4, 0], [3, 2], [1, 3]].
        """
        holding = ([[self._spades, self.spade_suit],
                    [self._hearts, self.heart_suit],
                    [self._diamonds, self.diamond_suit],
                    [self._clubs, self.club_suit]])
        holding.sort(reverse=True)
        return holding

    def long_suit(self, minimum_holding):
        """Return the first suit with minimum_holding number of cards."""
        suit = None
        if self.spades >= minimum_holding:
            suit = self.spade_suit
        elif self.hearts >= minimum_holding:
            suit = self.heart_suit
        elif self.diamonds >= minimum_holding:
            suit = self.diamond_suit
        elif self.clubs >= minimum_holding:
            suit = self.club_suit
        return suit

    @property
    def suit_shape(self):
        """Return a list of suits in decreasing order by holding."""
        shape = []
        for suit in self.ordered_holding:
            shape.append(suit[1])
        return shape

    @staticmethod
    def higher_ranking_suit(suit_one, suit_two):
        """Return the higher ranking of two suits."""
        if suit_one > suit_two:
            suit = suit_one
        else:
            suit = suit_two
        return suit

    def has_sequence(self, suit):
        """Return True if hand contains a three card sequence starting with
            an honour."""
        result = False
        sequences = ['AKQ', 'KQJ', 'QJT', 'JT9', 'T98']
        cards = [card for card in self.cards if card.suit == suit]
        if len(cards) >= 3:
            for index, card in enumerate(cards[2:]):
                triple = cards[index].rank + cards[index+1].rank + card.rank
                if triple in sequences:
                    result = True
                    break
                break
        return result

    def suit_length(self, suit):
        """Return the length of a suit or a dummy value."""
        if suit.is_suit:
            return self.suit_holding[suit]
        else:
            return -999

    @staticmethod
    def _call_name(level, suit):
        """Return a call name form level and suit."""
        call_name = ''.join([str(level), suit])
        return call_name

    def tracer(self, module, get_frame, trace_value='', display=False, trace_message=''):
        """Print function trace information."""
        if self.trace_module:
            source_file = module.replace('bfg_components.source.', '')
            if self.trace_module == source_file:
                display = True
        else:
            source_file = ''
        display_all = 0
        suppress_all = 0
        if isinstance(trace_value, Bid):
            call_id = trace_value.call_id
            name = trace_value.name
        else:
            call_id = ''
            name = ''
        if (display or display_all) and not suppress_all:
            hand = self.__str__()
            if hand != self.last_hand:
                if self.last_hand:
                    print('')
                print(hand)
                self._set_trace_hand()
                self.last_hand = hand

            line_number = get_frame.f_lineno
            function = get_frame.f_code.co_name
            if trace_message:
                trace_message = ''.join([', ', trace_message])
            output = f'{line_number:03d}, call id={call_id}, {source_file}, {function:.<40}, {name}{trace_message}'
            print(output)

    def get_attributes_from_hand(self, hand):
        """Set the attributes of this object from a hand instance."""
        for key, item in hand.__dict__.items():
            self.__dict__[key] = item



"""Expose the classes in the API."""

import gettext
import locale
import os

LOC = locale.getdefaultlocale()[0]
LOC = 'en_GB'

locales_exe = os.path.join(os.getcwd(), 'locale')
locales_run = '../bfgsupport/bfgsupport/locale'
if os.path.isdir(locales_exe):
    LOCALES_DIRECTORY = locales_exe
else:
    LOCALES_DIRECTORY = locales_run

if not os.path.isdir(LOCALES_DIRECTORY):
    LOCALES_DIRECTORY = '/locale'


translation = gettext.translation('bfg', localedir=LOCALES_DIRECTORY, languages=[LOC], fallback=False)
translation.install()
_ = translation.gettext


from ._version import __version__
VERSION = __version__

from .source.hand import Hand
from .source.comment_xref import CommentXref, convert_text_to_html
from .source.strategy_xref import StrategyXref, strategy_descriptions
from .source.bidding import Bid, Pass, Double
from .source.player import Player



""" Bid for Game
    Acol bidding module
"""

# trace 2.0.8

from bridgeobjects import ROLES
from .bidding import Bid, Pass
from .acol_openers_bid import OpeningBid
from .acol_openers_rebid import OpenersReBid
from .acol_openers_third_bid import OpenersThirdBid
from .acol_openers_later_bid import OpenersLaterBid
from .acol_responders_bid import RespondersBid
from .acol_responders_rebid import RespondersRebid
from .acol_responders_later_bid import RespondersLaterBid
from .acol_overcallers_bid import OverCallersBid
from .acol_overcallers_rebid import OverCallersRebid
from .acol_overcallers_third_bid import OverCallersThirdBid
from .acol_advancers_bid import AdvancersBid
from .acol_advancers_rebid import AdvancersRebid
from .acol_advancers_later_bid import AdvancersLaterBid


class AcolBid(object):
    """BfG AcolBid class."""
    def __init__(self, hand, board, role):
        """
            Identify the bidder and direct control accordingly.

            Return a bid.
        """
        self.board = board
        bid_history = board.active_bid_history
        self.bid = self.get_bid(hand.cards, bid_history, role)

    def get_bid(self, hand_cards, bid_history, role):
        """Direct bidding to correct bidder and return a bid."""
        if role == ROLES['Opener']:
            bid = self.openers_bid(hand_cards, bid_history)
        elif role == ROLES['Responder']:
            bid = self.responders_bid(hand_cards, bid_history)
        elif role == ROLES['Overcaller']:
            bid = self.overcallers_bid(hand_cards, bid_history)
        elif role == ROLES['Advancer']:
            bid = self.advancers_bid(hand_cards, bid_history)
        else:
            assert False, 'Bid not defined'
        return bid

    def openers_bid(self, hand_cards, bid_history):
        """Return opener's bid."""
        if not bid_history:
            bid = OpeningBid(hand_cards, self.board).suggested_bid()
        elif len(bid_history) == 4:
            bid = OpenersReBid(hand_cards, self.board).suggested_bid()
        elif len(bid_history) == 8:
            bid = OpenersThirdBid(hand_cards, self.board).suggested_bid()
        else:
            bid = OpenersLaterBid(hand_cards, self.board).suggested_bid()
        return bid

    def responders_bid(self, hand_cards, bid_history):
        """Return responder's bid."""
        if len(bid_history) == 2:
            bid = RespondersBid(hand_cards, self.board).suggested_bid()
        elif len(bid_history) == 6:
            bid = RespondersRebid(hand_cards, self.board).suggested_bid()
        else:
            bid = RespondersLaterBid(hand_cards, self.board).suggested_bid()
        return bid

    def overcallers_bid(self, hand_cards, bid_history):
        """Return overcaller's bid."""
        if len(bid_history) <= 4:
            bid = OverCallersBid(hand_cards, self.board).suggested_bid()
        elif len(bid_history) <= 8:
            bid = OverCallersRebid(hand_cards, self.board).suggested_bid()
        elif len(bid_history) <= 12:
            bid = OverCallersThirdBid(hand_cards, self.board).suggested_bid()
        else:
            bid = Pass('130')
        return bid

    def advancers_bid(self, hand_cards, bid_history):
        """Return overcaller responder's bid."""
        if (len(bid_history) <= 6 or
                (Bid(bid_history[-4]).is_pass and
                 len(bid_history) <= 6)):
            bid = AdvancersBid(hand_cards, self.board).suggested_bid()
        elif len(bid_history) <= 10:
            bid = AdvancersRebid(hand_cards, self.board).suggested_bid()
        elif len(bid_history) <= 14:
            bid = AdvancersLaterBid(hand_cards, self.board).suggested_bid()
        else:
            bid = Pass('130')
        return bid



""" Bid for Game
    Acol OverCallersBid module
"""

import inspect
from .bidding import Bid, Pass, Double
from .hand import Hand

import time


class OverCallersBid(Hand):
    """BfG OverCallersBid class."""
    def __init__(self, hand_cards, board):
        now = time.time()
        super(OverCallersBid, self).__init__(hand_cards, board)
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object"""
        now = time.time()
        suit = self.longest_suit
        if self._has_eight_points_and_eight_card_major():
            bid = self.bid_to_game(self.longest_suit, '0539')
        elif self._has_eight_points_and_eight_card_minor():
            bid = self.next_level_bid(suit, '0325')
        elif self._can_bid_seven_card_suit_over_openers_weak_two():
            bid = self.next_level_bid(self.longest_suit, '0252')
        elif len(self.bid_history) >= 5 and self.opener_bid_one.is_suit_call:
            bid = Pass('0540')
        elif self.opener_bid_one.name == '2C' or self.opener_bid_one.name == '2NT':
            bid = Pass('0541')
        elif self.hcp >= 15:
            bid = self._strong_bid()
        elif self.shape[0] >= 5 and self.hcp >= 9:
            bid = self._suit_overcall()
        elif self._can_make_weak_balanced_overcall():
            bid = self._weak_balanced_overcall()
        elif self._can_bid_six_card_suit_at_two_level():
            bid = self.next_level_bid(self.longest_suit, '0270')
        elif self._can_double_in_pass_out_seat():
            bid = Double('0501')
        else:
            bid = Pass('0542')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_bid(self):
        """Overcall with strong hand."""
        if self._very_strong_semi_balanced_with_stoppers():
            bid = self.nt_bid(3, '0543')
        elif self.hcp >= 19 and self.nt_level <= 2:
            bid = Double('0544')
        elif self._is_strong_and_semi_balanced_with_stoppers_over_weak_two():
            bid = self.nt_bid(2, '0545')
        elif self._is_strong_and_semi_balanced_with_stoppers_over_one_level():
            bid = self.nt_bid(self.nt_level, '0546')
        elif self._is_strong_shortage_in_bid_suits_and_minimum_in_unbid_suits():
            bid = Double('0547')
        elif self._is_strong_and_minimum_in_unbid_suits():
            bid = Double('0548')
        elif self._has_fifteen_points_and_minimum_in_openers_suit():
            bid = Double('0549')
        elif self._is_strong_and_opener_bid_one_nt():
            bid = Double('0550')
        elif self.is_balanced:
            bid = self._strong_bid_balanced()
        elif self.hcp >= 16 and self.shape == [4, 4, 4, 1] and self.nt_level <= 2:
            bid = self._strong_bid_4441()
        else:
            bid = self._suit_overcall()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_bid_4441(self):
        """Overcall with strong 4441 hand."""
        if self._two_or_fewer_in_openers_suit():
            bid = Double('0551')
        else:
            bid = self.next_nt_bid('0552')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_bid_balanced(self):
        """Overcall with strong balanced hand."""
        if 20 <= self.hcp <= 22 and self.stoppers_in_bid_suits and self.nt_level <= 2:
            bid = self.nt_bid(2, '0553')
        elif self._has_eighteen_points_and_stoppers_in_bid_suits():
            bid = Double('0554')
        elif self.hcp >= 15 and self.stoppers_in_bid_suits and self.nt_level == 1:
            bid = self.nt_bid(1, '0555')
        elif self._has_sixteen_points_and_shortage_or_opener_bid_one_nt():
            bid = Double('0556')
        else:
            bid = self._suit_overcall()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _suit_overcall(self):
        """Overcall with suit."""
        if self._has_five_card_biddable_suit():
            bid = self._overcall_with_suit()
        else:
            bid = Pass('0557')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcall_with_suit(self):
        """Return a bid in chosen suit."""
        if self.opener_bid_one.level == 1:
            bid = self._overcalls_at_one_level()
        elif self.opener_bid_one.level == 2:
            bid = self._overcall_after_weak_two()
        elif self.opener_bid_one.level == 3:
            bid = self._overcall_after_weak_three()
        else:
            bid = Pass('0558')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcalls_at_one_level(self):
        """Return  overcall bid after one level opening."""
        suit = self._select_suit()
        if self.hcp >= 16 and self.six_card_suit:
            raise_level = 1
            if self.responder_bid_one.is_value_call:
                raise_level = 0
            bid = self.next_level_bid(suit, '0559', raise_level)
        elif self._can_double():
            bid = Double('0560')
        elif self._has_twelve_points_and_void_in_bid_suits():
            bid = Double('0000')
        elif self.next_level(self._select_suit()) <= 1:
            bid = self._overcall_suit_at_one_level()
        elif self.next_level(self._select_suit()) <= 2:
            bid = self._overcall_suit_at_two_level()
        elif (self.hcp >= 15 and self.longest_suit not in self.opponents_suits and
              (self.next_level(self.longest_suit) <= 2 or
               self.last_bid.name != '2NT')):
            bid = self.next_level_bid(self.longest_suit, '0830')
        else:
            bid = Pass('0562')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcall_suit_at_one_level(self):
        """Overcall a suit at the 1 level."""
        suit = self._select_suit()
        if self.hcp > 20 and self.is_balanced:
            bid = self._strong_balanced_overcall()
        elif self._has_nine_points_and_some_shape(suit):
            bid = self.suit_bid(1, suit, '0563')
        elif self._can_bid_good_suit_at_one_level(suit):
            bid = self.suit_bid(1, suit, '0564')
        else:
            bid = Pass('0591')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcall_suit_at_two_level(self):
        """Overcall a suit at the 2 level."""
        suit = self._select_suit()
        suit_points = self.suit_points(suit)
        hand_points = self.hcp + self.distribution_points
        if self._has_strong_seven_card_suit():
            bid = self.suit_bid(3, suit, '0909')
        elif suit_points >= 9 and self.hcp >= 16:
            bid = Double('0565')
        elif self.shape == [5, 4, 4, 0] and self.hcp >= 11 and self.opener_bid_one.name != '1NT':
            bid = Double('0283')
        elif hand_points >= 12 and self.suit_points(self.longest_suit) >= 3:
            bid = self.suit_bid(2, suit, '0566')
        elif hand_points >= 12 and self.five_five_or_better:
            bid = self.suit_bid(2, suit, '0567')
        elif hand_points >= 12 and self.shape[0] >= 6:
            bid = self.suit_bid(2, suit, '0568')
        elif hand_points >= 12 and self.hcp + suit_points >= 17:
            bid = self.suit_bid(2, suit, '0569')
        else:
            bid = self._strong_balanced_overcall()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcall_after_weak_two(self):
        """Return overcall bids after weak two opening."""
        suit = self._select_suit()
        if self.shape[0]+self.suit_points(suit) >= 9:
            bid = self._overcall_suit_after_weak_two()
        elif self._has_twelve_points_shortage_and_minimum_in_unbid_suits():
            bid = Double('0572')
        else:
            bid = Pass('0570')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcall_suit_after_weak_two(self):
        """Return overcall suit after weak two opening."""
        suit = self._select_suit()
        if self._can_bid_suit_at_level_two(suit):
            bid = self.suit_bid(2, suit, '0571')
        elif self.hcp >= 16:
            bid = Double('0573')
        elif self.next_level(suit) == 3 and self.hcp >= 12:
            bid = self.suit_bid(3, suit, '0574')
        # elif self._has_thirteen_points_and_can_bid_six_card_suit_at_level_three(suit):
        #     bid = self.next_level_bid(suit, '0575')
        elif self._has_six_card_suit_can_overcall_weak_two():
            bid = self.next_level_bid(self.longest_suit, '0672')
        else:
            bid = Pass('0576')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcall_after_weak_three(self):
        """Overcall after weak three."""
        if (self.shape == [5, 4, 4, 0] and
                self.suit_length(self.opener_bid_one.denomination) == 0):
            bid = Double('0915')
        elif (self.hcp >= 19 and
                self.suit_length(self.opener_bid_one.denomination) <= 2 and
                self.four_card_major):
            bid = Double('0518')
        elif self._can_rebid_over_weak_three():
            bid = self._overcall_suit_after_weak_three()
        else:
            bid = Pass('0577')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _overcall_suit_after_weak_three(self):
        """Return overcall suit after weak three opening."""
        suit = self._select_suit()
        if self._can_bid_nt_with_minor_suit(suit):
            bid = self.nt_bid(3, '0578')
        elif self._has_strong_hand(suit) and self.opponents_at_game:
            bid = Double('0579')
        elif self._has_strong_hand(suit) and self.next_level(suit) <= 4:
            bid = self.next_level_bid(suit, '0580')
        elif self.next_level(suit) <= 3:
            bid = self.suit_bid(3, suit, '0581')
        elif self._can_bid_suit_at_level_four(suit):
            bid = Double('0582')
        else:
            bid = Pass('0583')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _weak_balanced_overcall(self):
        """Overcall with a weak balanced hand."""
        if self._opponents_not_in_fit_at_level_one():
            bid = Double('0584')
        elif self._fffo_and_opener_bids_suit():
            bid = Double('0350')
        elif self._can_double_weak_two():
            bid = Double('0850')
        else:
            bid = Pass('0586')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_balanced_overcall(self):
        """Overcall with a strong balanced hand."""
        if self.is_balanced:
            bid = self._strong_balanced_overcall_make_bid()
        else:
            bid = Pass('0587')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_balanced_overcall_make_bid(self):
        """Overcall with a strong balanced hand."""
        if self._fewer_than_seventeen_points_can_bid_one_nt():
            bid = self.nt_bid(1, '0588')
        elif self._eighteen_plus_points_stoppers_minimum_in_unbid_suits():
            bid = Double('0589')
        elif self.hcp >= 15 and self.shape[0] >= 5:
            bid = self.next_level_bid(self.longest_suit, comment='0590')
        else:
            bid = Pass('0303')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    # Various utility functions

    def _has_strong_hand(self, suit):
        """Return True if hand is strong with good suit."""
        suit_points = self.suit_points(suit)
        return (self.hcp >= 16 and
                self.shape[0] >= 5 and
                suit_points >= 6)

    def _has_minimum_hand(self, suit):
        """Return True if hand is minimum with good suit."""
        suit_points = self.suit_points(suit)
        return ((self.hcp >= 14 and
                 self.shape[0] >= 5 and
                 suit_points >= 6) or
                (self.hcp >= 12 and
                 self.shape[0] >= 6 and
                 suit_points >= 5) or
                (self.hcp >= 11 and
                 self.shape[0] >= 7 and
                 suit_points >= 4))

    def _select_suit(self):
        """Return selected overcall suit."""
        if 5 <= self.shape[0] == self.shape[1]:
            long_suit = self.higher_ranking_suit(self.longest_suit,
                                                 self.second_suit)
        elif self.shape[0] >= 5:
            long_suit = self.longest_suit
        else:
            long_suit = self.longest_suit
        chosen_suit = long_suit
        if long_suit in self.opponents_suits:
            chosen_suit = self._second_five_card_suit(long_suit)
        return chosen_suit

    def _second_five_card_suit(self, long_suit):
        """Select 5 card suit if long suit is opponent's suit."""
        if self.spades >= 5 and long_suit != self.spade_suit:
            suit = self.spade_suit
        elif self.hearts >= 5 and long_suit != self.heart_suit:
            suit = self.heart_suit
        elif self.diamonds >= 5 and long_suit != self.diamond_suit:
            suit = self.diamond_suit
        elif self.clubs >= 5 and long_suit != self.club_suit:
            suit = self.club_suit
        else:
            suit = self.no_trumps
        return suit

    def _minimum_in_unbid_suits(self):
        """
            Return True if hand has the minimum required number of
            cards in each unbid suit.
        """
        value = True
        if not self.overcaller_in_fourth_seat:
            minimum = 3
        elif self.responder_bid_one.is_pass:
            minimum = 3
        else:
            minimum = 4
        bid_suits = [Bid(bid).denomination for bid in self.bid_history[::2]]
        for suit in self.suits:
            if suit not in bid_suits:
                if self.suit_length(suit) <= minimum - 1:
                    value = False
                    break
        # if self.opener_bid_one.denomination.name == 'S':
        #     if self.opener_bid_one.level == 1 and self.suit_length(self.heart_suit) < 4:
        #         value = False
        return value

    def _two_or_fewer_in_openers_suit(self):
        """Return True if hand holds two or fewer in opener's suit."""
        return self._fewer_in_openers_suit(2)

    def _three_or_fewer_in_openers_suit(self):
        """Return True if hand holds three or fewer in opener's suit."""
        return self._fewer_in_openers_suit(3)

    def _fewer_in_openers_suit(self, limit):
        """Return True if hand holds fewer than limit in opener's suit."""
        value = False
        if self.opener_bid_one.is_suit_call:
            if self.suit_length(self.opener_bid_one.denomination) <= limit:
                value = True
        if self.responder_bid_one.is_suit_call:
            value = False
            if self.suit_length(self.responder_bid_one.denomination) <= limit:
                value = True
        return value

    def _three_minimum(self):
        """Return True if hand is void or 3+ cards in each suit."""
        value = True
        for suit_holding in self.shape:
            if 1 <= suit_holding <= 2:
                value = False
                break
        return value

    def _void_in_bid_suits(self):
        """Return True if hand is void in bid suit."""
        value = False
        bid_suits = []
        for bid in self.bid_history[::2]:
            suit = Bid(bid).denomination
            if suit.is_suit:
                bid_suits.append(suit)
        for suit in bid_suits:
            if self.suit_length(suit) == 0:
                value = value and True
            else:
                value = False
        return value

    # Various boolean functions

    def _has_eight_points_and_eight_card_major(self):
        """Return True if pre-emptive 8 card suit."""
        result = (self.shape[0] >= 8 and
                  self.hcp >= 8 and
                  self.longest_suit not in self.opponents_suits and
                  self.next_level(self.longest_suit) <= self.longest_suit.game_level and
                  self.longest_suit.is_major)
        return result

    def _has_eight_points_and_eight_card_minor(self):
        """Return True if zzz."""
        result = (self.shape[0] >= 8 and
                  self.hcp >= 8 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _can_bid_seven_card_suit_over_openers_weak_two(self):
        """Return True if can bid 7 card suit over weak two opening."""
        result = (self.opener_bid_one.level == 2 and
                  self.shape[0] >= 7 and
                  self.hcp >= 12 and
                  self.longest_suit not in self.opponents_suits and
                  self.next_level(self.longest_suit) <= 2)
        return result

    def _can_make_weak_balanced_overcall(self):
        """Return True if can make weak balanced overcall."""
        result = (self.hcp >= 12 and
                  ((self.suit_length(self.opener_bid_one.denomination) <= 2 or
                    self.opener_bid_one.is_nt) or
                   (self.shape == [4, 3, 3, 3] and len(self.bid_history) >= 2)))
        return result

    def _can_bid_six_card_suit_at_two_level(self):
        """Return True if can bid a six card suit at the two level."""
        result = (self.shape[0] >= 6 and
                  self.hcp >= 8 and
                  self.longest_suit.is_major and
                  self.longest_suit not in self.opponents_suits and
                  self.opener_bid_one.level == 1 and
                  self.next_level(self.longest_suit) <= 2)
        return result

    def _very_strong_semi_balanced_with_stoppers(self):
        """Return True if very strong and semi_balanced."""
        result = (self.hcp >= 22 and
                  self.is_semi_balanced and
                  self.has_stopper(self.opener_bid_one.denomination) and
                  self.nt_level <= 3)
        return result

    def _is_strong_and_semi_balanced_with_stoppers_over_weak_two(self):
        """Return True if strong and semi-balanced over weak two."""
        result = (self.hcp >= 16 and
                  self.is_semi_balanced and
                  self.opener_bid_one.level == 2 and
                  self.has_stopper(self.opener_bid_one.denomination) and
                  self.nt_level <= 2)
        return result

    def _is_strong_and_semi_balanced_with_stoppers_over_one_level(self):
        """Return True if zzz."""
        result = (16 <= self.hcp <= 17 and
                  self.is_balanced and
                  self.opener_bid_one.is_suit_call and
                  self.has_stopper(self.opener_bid_one.denomination) and
                  self.nt_level <= 1)
        return result

    def _is_strong_shortage_in_bid_suits_and_minimum_in_unbid_suits(self):
        """Return True if strong and has minimum in unbid suits."""
        result = (self.hcp >= 16 and
                  self._two_or_fewer_in_openers_suit() and
                  self._minimum_in_unbid_suits() and
                  self.responder_bid_one.is_pass and
                  not self.previous_bid.is_game)
        return result

    def _is_strong_and_minimum_in_unbid_suits(self):
        """Return True if strong and minimum in unbid suits."""
        result = (self.hcp >= 16 and self._two_or_fewer_in_openers_suit() and
                  self._minimum_in_unbid_suits() and
                  not self.previous_bid.is_game)
        return result

    def _has_fifteen_points_and_minimum_in_openers_suit(self):
        """Return True if 15 points and fewer than 2 in openers suit."""
        result = (self.hcp >= 15 and
                  self._two_or_fewer_in_openers_suit() and
                  self._minimum_in_unbid_suits() and
                  not self.previous_bid.is_game and
                  self.next_level(self.longest_suit) > 1 and
                  not (self.shape[0] >= 5 and
                       self.suit_points(self.longest_suit) >= 4))
        return result

    def _is_strong_and_opener_bid_one_nt(self):
        """Return True if 16 points and opener has bid 1NT."""
        result = (self.hcp >= 16 and
                  self.opener_bid_one.name == '1NT' and
                  not self.responder_bid_one.is_value_call)
        return result

    def _has_eighteen_points_and_stoppers_in_bid_suits(self):
        """Return True if 18/19 points and stoppers in bid suits."""
        result = (18 <= self.hcp <= 19 and
                  self.stoppers_in_bid_suits and
                  self.longest_suit not in self.opponents_suits and
                  self.nt_level <= 2)
        return result

    def _has_sixteen_points_and_shortage_or_opener_bid_one_nt(self):
        """Return True if zzz."""
        result = (self.hcp >= 16 and
                  (self._three_or_fewer_in_openers_suit() or
                   self.opener_bid_one.is_nt))
        return result

    def _has_five_card_biddable_suit(self):
        """Return True if has 5 card biddable suit."""
        chosen_suit = self._select_suit()
        result = (chosen_suit.is_suit and
                  self.suit_length(chosen_suit) >= 5 and
                  chosen_suit not in self.opponents_suits)
        return result

    def _has_twelve_points_and_void_in_bid_suits(self):
        """Return True if 23 points and void in bid suits and minimum in bid suits."""
        result = (self.hcp >= 12 and
                  self._void_in_bid_suits() and
                  self._three_minimum() and
                  not self.five_five and
                  not self.shape[0] >= 6 and
                  self.nt_level <= 3)
        return result

    def _has_nine_points_and_some_shape(self, suit):
        """Return True if 9/10 points and shapely."""
        hand_points = self.hcp + self.distribution_points
        result = (9 <= hand_points <= 10 and
                  self.suit_points(suit) >= 6 and
                  self.next_level(suit) == 1 or
                  self.five_five or
                  (self.shape[0] >= 6 and
                   self.shape[0]+self.suit_points(suit) >= 9))
        return result

    def _can_bid_good_suit_at_one_level(self, suit):
        """Return True if can bid a strong suit at one level."""
        hand_points = self.hcp + self.distribution_points
        result = (hand_points >= 11 and
                  self.next_level(suit) == 1 and
                  (self.suit_points(suit) >= 1 or
                   (self.suit_points(suit) >= 1 and self.has_sequence(suit))))
        return result

    def _has_strong_seven_card_suit(self):
        """Return True if has strong 7 card suit."""
        result = (self.hcp + self.distribution_points >= 13 and
                  self.shape[0] >= 7 and
                  self.suit_points(self.longest_suit) == 10)
        return result

    def _can_bid_suit_at_level_two(self, suit):
        """Return True if 12 points or seven card suit and level <= 2."""
        result = (self.next_level(suit) <= 2 and
                  (self.shape[0] >= 7 or
                   (self.shape[0] >= 6 and self.hcp >= 11) or
                   (self.shape[0] == 5 and self.hcp >= 12)))
        return result

    def _has_twelve_points_shortage_and_minimum_in_unbid_suits(self):
        """Return True if 12 points shortage and minimum in unbid suits."""
        result = (self.hcp >= 12 and
                  self.suit_length(self.opener_bid_one.denomination) <= 2 and
                  # self._minimum_in_unbid_suits() and
                  (self.shape == [4, 4, 3, 2] or
                   self.shape == [5, 4, 3, 1]))
        return result

    # def _has_thirteen_points_and_can_bid_six_card_suit_at_level_three(self, suit):
    #     """Return True if 13 points and can bid 6 card suit at 3 level."""
    #     result = (self.hcp >= 13 and
    #               self.shape[0] >= 6 and
    #               self.next_level(suit) <= 3)
    #     return result

    def _can_rebid_over_weak_three(self):
        """Return True if hand suitable for overcalling a weak 3."""
        suit = self._select_suit()
        result = self._has_strong_hand(suit) or self._has_minimum_hand(suit)
        return result

    def _can_bid_nt_with_minor_suit(self, suit):
        """Return True if suit is minor can bid NT."""
        result = (suit.is_minor and
                  self.suit_points(suit) >= 9 and
                  self.stoppers_in_bid_suits and
                  self.nt_level <= 3)
        return result

    def _can_bid_suit_at_level_four(self, suit):
        """Return True if can bid suit ay level 4."""
        result = (self.suit_length(self.opener_bid_one.denomination) <= 2 and
                  self._minimum_in_unbid_suits() and
                  self.next_level(suit) <= 4)
        return result

    def _opponents_not_in_fit_at_level_one(self):
        """Return True if opponents bid different suits at level 1."""
        result = (self.opener_bid_one.level <= 1 and
                  self._minimum_in_unbid_suits() and
                  self.responder_bid_one.level <= 1 and
                  self.opener_bid_one.is_suit_call and
                  not self.responder_bid_one.is_nt and
                  not self.opponents_at_game)
        return result

    def _fffo_and_opener_bids_suit(self):
        """Return True if strong or 4441."""
        result = (self.four_four_four_one and
                  self.opener_bid_one.is_suit_call and
                  not self.opponents_at_game)
        return result

    def _fewer_than_seventeen_points_can_bid_one_nt(self):
        """Return True if fewer then 17 points stopper and nt_level 1."""
        result = (15 <= self.hcp <= 17 and
                  self.stoppers_in_bid_suits and
                  self.nt_level == 1)
        return result

    def _eighteen_plus_points_stoppers_minimum_in_unbid_suits(self):
        """Return True if 18+ points, stoppers and minimum in bid suits."""
        result = (self.hcp >= 18 and
                  self.stoppers_in_bid_suits and
                  self._minimum_in_unbid_suits())
        return result

    def _can_double_in_pass_out_seat(self):
        """Return True if can double in the pass out seat."""
        result = (len(self.bid_history) == 3 and
                  self.hcp >= 11 and
                  self.four_four_four_one and
                  self.suit_length(self.opener_bid_one.denomination) == 1)
        return result

    def _has_six_card_suit_can_overcall_weak_two(self):
        """Return True if can overcall weak two."""
        result = (self.hcp >= 10 and
                  # self.suit_points(self.opener_bid_one.denomination) >= 7 and
                  self.shape[0] >= 6 and
                  self.suit_points(self.longest_suit) >= 3 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _can_double(self):
        """Return True if double is allowed."""
        can_double = False
        if len(self.bid_history) == 1:
            can_double = True
        elif (len(self.bid_history) == 3 and
                self.responder_bid_one.is_value_call and
                self.opener_bid_one.denomination == self.responder_bid_one.denomination):
            can_double = True

        result = ((self.hcp >= 16 or (self.hcp >= 15 and self.shape[0] >= 7)) and can_double)
        return result

    def _can_double_weak_two(self):
        """Return True if can double a weak two."""
        result = (self.hcp >= 12 and
                  self.shape[2] >= 3 and
                  self.suit_length(self.opener_bid_one.denomination) <= 2 and
                  self.opener_bid_one.level ==2 and
                  self.opener_bid_one.denomination.name != 'S')
        return result

    # def xxx(self):
    #     """Return True if zzz."""
    #     result = False
    #     return result

    # def xxx(self):
    #     """Return True if zzz."""
    #     result = False
    #     return result



""" Bid for Game
    Acol OverCallersBid module
"""

import inspect
from .bidding import Bid, Pass, Double
from .hand import Hand


class OverCallersRebid(Hand):
    """BfG OverCallersRebid class."""
    def __init__(self, hand_cards, board):
        super(OverCallersRebid, self).__init__(hand_cards, board)
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        if self._can_double():
            bid = Double('0692')
        elif self._very_strong_six_card_suit_advancer_bids_game():
            bid = self.suit_bid(6, self.longest_suit, '0693')
        elif self.bid_one.is_pass:
            bid = Pass('0694')
        elif self._advancer_bids_three_nt_and_long_suit():
            bid = self.bid_to_game(self.longest_suit, '0311')
        elif self.advancer_bid_one.is_game:
            bid = Pass('0695')
        elif self._advancer_supports_overcaller_suit():
            bid = self._partner_shows_support()
        elif self.advancer_bid_one.is_nt and self.is_semi_balanced:
            bid = self._bid_after_nt_agreed()
        elif self.advancer_bid_one.is_suit_call:
            bid = self._partner_bid_suit()
        elif self.bid_one.is_double:
            bid = self._rebid_after_double_pass()
        elif self.hcp >= 15 and self.shape[0] >= 6:
            bid = self._rebid_long_suit()
        elif self.advancer_bid_one.is_value_call and self.shape[0] >= 7:
            bid = self._rebid_long_suit()
        elif self._advancer_bids_two_nt_and_six_card_suit_and_weak():
            bid = self.next_level_bid(self.longest_suit, '0462')
        elif self._strong_over_weak_three():
            bid = Double('0691')
        elif (self.opener_bid_one.is_suit_call and
              self.opener_bid_one.denomination == self.responder_bid_one.denomination):
            bid = Pass('0696')
        elif self.hcp >= 13 and self.five_five:
            bid = self._strong_five_five_hand()
        elif self._has_long_suit():
            bid = self._rebid_long_suit()
        elif self.shape[0] >= 6:
            bid = self._six_card_suit_bid()
        elif self._can_bid_second_suit_at_level_two():
            bid = self.suit_bid(2, self.second_suit, '0652')
        elif self._can_rebid_longest_suit_at_level_two():
            bid = self.next_level_bid(self.longest_suit, '0697')
        elif self._can_rebid_second_suit_at_level_two():
            bid = self.next_level_bid(self.second_suit, '0698')
        elif self._can_bid_second_suit():
            bid = self.next_level_bid(self.second_suit, '0569')
        else:
            bid = Pass('0699')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _bid_after_nt_agreed(self):
        """Bid after NT agreed."""
        if self._strong_with_biddable_five_card_major():
            bid = self._bid_five_card_major()
        elif self.nt_level <= 3 and self.hcp >= 17:
            bid = self.nt_bid(3, '0700')
        elif self.hcp <= 14 and self.shape[0] >= 6:
            bid = self.next_level_bid(self.longest_suit, '0701')
        elif self._can_rebid_nt_at_three_level():
            bid = self.nt_bid(3, '0702')
        elif self._can_rebid_second_suit_at_level_two():
            raise_level = 0
            suit = self.second_suit
            if self.second_suit in self.opponents_suits:
                suit = self.longest_suit
                if self.hcp >= 14:
                    raise_level = 1
            bid = self.next_level_bid(suit, '0703', raise_level)
        else:
            bid = Pass('0704')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _bid_five_card_major(self):
        """Return bid after NT agreed with 5 card major."""
        suit = self._biddable_five_card_major()
        if self._is_competitive_auction():
            if self.hcp >= 19 and self.next_level(suit) <= 4:
                bid = self.suit_bid(4, suit, '0705')
            elif self.next_level(suit) <= 3:
                bid = self.suit_bid(3, suit, '0036')
            else:
                bid = Pass('0000')
        elif self._strong_and_advancer_bids_two_nt():
            bid = self.nt_bid(3, '0047')
        elif self.next_level(suit) <= 2:
            bid = self.suit_bid(2, suit, '0037')
        else:
            bid = Pass('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_five_five_hand(self):
        """Bid with strong  5/5 hand."""
        suit = self._five_five_second_suit()
        bid = self.next_level_bid(suit, '0708')
        if bid.level >= 4:
            bid = Pass('0709')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _six_card_suit_bid(self):
        """Bid with a 6+ card suit."""
        if self._can_show_second_suit_after_advancer_has_bid():
            bid = self.next_level_bid(self.second_suit, '0710')
        elif self._can_bid_six_card_suit_in_competitive_auction():
            bid = self.next_level_bid(self.longest_suit, '0711')
        elif self.advancer_bid_one.is_nt:
            bid = self.next_level_bid(self.bid_one.denomination, '0712')
        else:
            bid = Pass('0713')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_long_suit(self):
        """Rebid own long suit."""
        if self._can_rebid_suit_after_advancer_has_bid():
            bid = self.next_level_bid(self.bid_one.denomination, '0714')
        elif self.shape[0] >= 7 and self.next_level(self.longest_suit) <= 2:
            bid = self.next_level_bid(self.longest_suit, '0937')
        else:
            bid = Pass('0715')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_double(self):
        """Rebid after double."""
        suit_for_rebid = self._get_best_suit_for_rebid()
        if self._weak_hand_after_opener_bids_weak_two():
            bid = Pass('0716')
        elif self._minimum_after_double_over_1nt():
            bid = Pass('0857')
        elif self.is_balanced and self.hcp >= 22 and self.nt_level <= 3:
            bid = self._exceptional_major_or_nt()
        elif self.hcp >= 18 and self.nt_level <= 2:
            bid = self._rebid_after_double_strong()
        elif self._is_strong_in_opponents_suit():
            bid = Pass('0402')
        elif self._strong_and_no_support_for_advancer(suit_for_rebid):
            bid = self.next_level_bid(suit_for_rebid, '0717')
        elif self._very_strong_and_can_bid_nt():
            bid = self.next_nt_bid('0718')
        elif self._can_support_advancer_after_double():
            bid = self._no_long_suit_support_partner()
        elif self.suit_length(self.advancer_bid_one.denomination) >= 4:
            bid = self.next_level_bid(self.advancer_bid_one.denomination, '0719')
        elif self._suit_for_rebid_is_minor_after_double(suit_for_rebid):
            bid = self.next_level_bid(self.advancer_bid_one.denomination, '0720')
        elif self._strong_and_can_bid_rebid_suit(suit_for_rebid):
            bid = self.next_level_bid(suit_for_rebid, '0721')
        else:
            bid = self._rebid_suit_for_rebid()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _exceptional_major_or_nt(self):
        """Bid major or NT on very exceptional hand."""
        if self._can_bid_five_card_major_at_three_level():
            bid = self.suit_bid(3, self.longest_suit, '0722')
        else:
            bid = self.nt_bid(3, '0723')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_long_suit_support_partner(self):
        """Bid with no long suit and 3+ card support for partner."""
        if self.advancer_bid_one.level >= 4:
            if self.hcp + self.distribution_points >= 21:
                bid = self.next_level_bid(self.advancer_bid_one.denomination, '0724')
            else:
                bid = Pass('0725')
        elif self._has_opponents_suit(self.bid_history[-1]):
            bid = Pass('0726')
        elif self._strong_balanced_and_advancer_bids_minor():
            bid = self.nt_bid(3, '0246')
        elif self._above_minimum_and_advancer_bids_after_double():
            if self._can_raise_advancer_after_double():
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(self.advancer_bid_one.denomination, '0727', raise_level)
        elif self._support_for_partners_level_three_call():
            bid = self.next_level_bid(self.partner_bid_one.denomination, '0784')
        else:
            bid = Pass('0728')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_suit_for_rebid(self):
        """Bid with weak hand and no support for/from partner."""
        if self.shape[0] >= 5:
            suit_for_rebid = self.longest_suit
        else:
            suit_for_rebid = self._cheapest_four_card_suit()
        if self._partner_weak_after_double():
            bid = Pass('0668')
        elif self._is_balanced_with_stopper(suit_for_rebid):
            bid = self.nt_bid(self.nt_level, '0729')
        elif suit_for_rebid not in self.opponents_suits:
            chosen_suit = self._select_suit_for_rebid()
            bid = self.next_level_bid(chosen_suit, '0730')
        else:
            bid = Pass('0731')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_double_strong(self):
        """Rebid after Double with a strong hand."""
        suit_for_rebid = self._get_best_suit_for_rebid()
        if self._four_card_support_for_advancers_major():
            bid = self._support_partners_major(self.advancer_bid_one.denomination)
        elif self._can_bid_five_card_major_at_two_level(suit_for_rebid):
            if suit_for_rebid not in self.opponents_suits:
                bid = self.suit_bid(2, suit_for_rebid, '0732')
            else:
                bid = self.next_nt_bid('0733')
        elif self._very_strong_and_some_support_for_partner():
            bid = self.bid_to_game(self.partner_last_bid.denomination, '0572')
        elif self._has_five_card_suit_and_no_support_from_advancer(suit_for_rebid):
            bid = self._five_card_suit_and_no_support_from_advancer()
        else:
            bid = self.next_nt_bid('0736')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _five_card_suit_and_no_support_from_advancer(self):
        if self.longest_suit in self.opponents_suits:
            suit_for_rebid = self.second_suit
        else:
            suit_for_rebid = self.longest_suit
        if self._is_balanced_and_long_suit_is_minor(suit_for_rebid):
            bid = self.next_nt_bid('0260')
        elif self._has_three_card_support_for_advancers_major(suit_for_rebid):
            bid = self._support_partners_major(self.advancer_bid_one.denomination)
        elif suit_for_rebid not in self.opponents_suits:
            bid = self.next_level_bid(suit_for_rebid, '0734')
        else:
            bid = Pass('0735')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_partners_major(self, partners_suit):
        """Bid with support for partners major."""
        level = self.next_level(partners_suit)
        if self.opener_bid_two.is_value_call:
            previous_bid = self.opener_bid_two
        elif self.responder_bid_one.is_value_call:
            previous_bid = self.responder_bid_one
        elif self.opener_bid_one.is_value_call:
            previous_bid = self.opener_bid_one
        else:
            assert False, 'previous_bid not assigned'
        if self.is_jump(previous_bid, self.advancer_bid_one):
            if level < 4:
                level = 4
            comment = '0737'
        else:
            comment = '0738'
        bid = self.suit_bid(level, partners_suit, comment)
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _rebid_after_double_pass(self):
        """Rebid after overcall Double and partner has passed."""
        suit = Bid(self.bid_history[-3]).denomination
        if self.bid_history[-1] == 'P' and self.bid_history[-2] == 'P' and self.suit_length(suit) >= 4 and self.suit_points(suit) >= 6:
            bid = Pass('0930')
        elif self._strong_and_opposition_weak():
            bid = Double('0740')
        elif self._is_strong_has_five_five():
            bid = self._strong_with_five_five()
        elif self._is_strong_with_good_five_card_suit():
            bid = self.next_level_bid(self.longest_suit, '0739')
        elif self.advancer_bid_one.is_double and not self.opponents_at_game:
            bid = self.next_level_bid(self.longest_suit, '0739')
        elif self._is_strong_and_balanced():
            bid = self.nt_bid(2, '0743')
        elif self._is_medium_and_balanced():
            bid = self.nt_bid(2, '0744')
        else:
            bid = Pass('0745')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _strong_with_five_five(self):
        if self.longest_suit in self.opponents_suits:
            suit_for_rebid = self._five_five_second_suit()
        else:
            suit_for_rebid = self.longest_suit
        if self.suit_points(suit_for_rebid) >= 5:
            bid = self.next_level_bid(suit_for_rebid, '0741')
        else:
            bid = Pass('0742')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _partner_bid_suit(self):
        """Respond after partner has bid a suit."""
        if self.bid_one.is_double:
            bid = self._rebid_after_double()
        elif self._advancer_bids_stayman():
            bid = self._response_to_stayman()
        elif self._can_bid_sevel_card_suit_at_level_three():
            bid = self.next_level_bid(self.longest_suit, '0795')
        elif self.suit_length(self.advancer_bid_one.denomination) >= 3:
            bid = self._support_for_partner()
        elif self.is_balanced and self.stoppers_in_bid_suits:
            bid = self._balanced_rebid()
        elif self.shape[0] >= 7 and self.advancer_bid_one.level == 3:
            bid = self.next_level_bid(self.bid_one.denomination, '0747')
        elif self._has_long_strong_suit():
            bid = self.next_level_bid(self.bid_one.denomination, '0748')
        elif self._has_five_four_and_can_bid_second_suit():
            bid = self.next_level_bid(self.second_suit, '0746')
        elif self._can_show_four_card_suit_at_level_two():
            bid = self.next_level_bid(self.second_suit, '0940')
        elif self._can_show_four_card_suit_at_level_four():
            bid = self.next_level_bid(self.second_suit, '0368')
        elif self._has_six_card_suit_and_no_support_for_advancer():
            if self.shape[0] >= 7:
                level = 1
            elif ((self.suit_length(self.advancer_bid_one.denomination) >= 2 and
                    self.advancer_bid_one.denomination.rank > self.bid_one.denomination.rank) or
                    self.hcp <= 12):
                level = 0
            else:
                level = 1
            bid = self.next_level_bid(self.bid_one.denomination, '0749', raise_level=level)
        elif self._can_bid_five_five_at_level_two():
            bid = self.next_level_bid(self.second_suit, '0750')
        elif self._can_bid_seven_card_suit_at_level_three():
            bid = self.next_level_bid(self.bid_one.denomination, '0751')
        elif self.my_last_bid.is_nt:
            bid = Pass('0744')
        elif self._has_twelve_points_no_support_for_advancer():
            bid = self.next_nt_bid('0752')
        elif self._good_second_suit_and_rebiddable_long_suit():
            bid = self.next_level_bid(self.second_suit, '0920')
        elif self.shape[0] >= 6 and self.longest_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.longest_suit, '0000')
        else:
            bid = Pass('0753')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _partner_shows_support(self):
        """Bid after partner has shown support."""
        hvp = self.hcp + self.distribution_points
        if self.hcp >= 19:
            bid = self.next_level_bid(self.bid_one.denomination, '0754', raise_level=1)
        elif hvp >= 16 and self.bid_one.denomination.is_major:
            bid = self.next_level_bid(self.bid_one.denomination, '0755')
        else:
            bid = self._partner_shows_support_weak_hand()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _partner_shows_support_weak_hand(self):
        """Bid after partner has shown support with weak hand."""
        if self._advancer_has_jumped_not_competitive():
            bid = self._partner_shows_support_with_jump()
        elif self._advancer_has_bid_at_level_three():
            bid = self.suit_bid(4, self.bid_one.denomination, '0756')
        elif self._advancer_has_jumped_over_openers_rebid():
            bid = self.next_level_bid(self.bid_one.denomination, '0758')
        elif self.six_four and self.next_level(self.bid_one.denomination) <= 3 and self.hcp >= 11:
            bid = self.next_level_bid(self.bid_one.denomination, '0759')
        elif (self._has_thirteen_points_and_can_bid_five_card_suit_at_level_three() and
                not self.bid_history[-1] == 'P'):
            bid = self.next_level_bid(self.bid_one.denomination, '0760')
        elif self._has_eleven_points_and_can_bid_seven_card_suit_at_level_three():
            bid = self.next_level_bid(self.bid_one.denomination, '0761')
        elif self._has_fifteen_points_can_bid_major():
            bid = self.next_level_bid(self.bid_one.denomination, '0762')
        else:
            bid = Pass('0763')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _partner_shows_support_with_jump(self):
        """Bid after partner has shown support with a jump."""
        level = self.next_level(self.bid_one.denomination)
        level += 1
        if self.bid_one.denomination.is_major:
            level = min(4, level)
        else:
            level = min(5, level)
        if self._can_raise_after_advancer_jumps(level):
            bid = self.suit_bid(level, self.bid_one.denomination, '0764')
        else:
            bid = Pass('0765')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _support_for_partner(self):
        """Bid with support for partner."""
        level = self._get_level_with_support_for_advancer()
        if self._advancer_has_bid_clubs_and_can_support(level):
            bid = self.suit_bid(level, self.advancer_bid_one.denomination, '0766')
        else:
            bid = Pass('0767')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _response_to_stayman(self):
        """Bid after partner has used Stayman"""
        if self.hearts >= 4 and self.heart_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.heart_suit, '0768')
        elif self.spades >= 4 and self.spade_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.spade_suit, '0769')
        elif self.next_level(self.diamond_suit) <= 2:
            bid = self.diamond_bid(2, '0237')
        elif self.nt_level <= 3:
            bid = self.nt_bid(3, '0770')
        else:
            bid = Pass('0000')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _no_support_for_partner(self, suit):
        """Bid with no support for partner."""
        level = self.next_level(suit)
        if level <= self.suit_length(suit) or self.advancer_bid_one.level == 3:
            bid = self.suit_bid(level, suit, '0771')
        else:
            bid = Pass('0772')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _balanced_rebid(self):
        """Bid with balanced hand."""
        if self.my_last_bid.is_nt:
            bid = self._balanced_overcall_nt()
        else:
            bid = self._balanced_rebid_after_suit()
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _balanced_overcall_nt(self):
        """Return bid with a balanced hand after overcaller bid NT."""
        if (self.advancer_bid_one.denomination == self.club_suit and
                Bid(self.bid_history[-1]).is_pass):
            bid = self._response_to_stayman()
        elif (self.my_last_bid.level >= 2 and
              self.suit_length(self.advancer_bid_one.denomination) >= 2):
            bid = self.next_level_bid(self.advancer_bid_one.denomination, '0773')
        else:
            bid = Pass('0774')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    def _balanced_rebid_after_suit(self):
        """Rebid with balanced hand after suit bid."""
        if (self.advancer_bid_one.level == 3 and
                self.hcp >= 14 and
                self.nt_level <= 3):
            bid = self.nt_bid(3, '0775')
        elif self.hcp >= 14 and self.nt_level <= 2:
            bid = self.nt_bid(2, '0776')
        else:
            bid = Pass('0777')
        self.tracer(__name__, inspect.currentframe(), bid, self.trace)
        return bid

    # Various utility functions

    def _cheapest_four_card_suit(self):
        """Return the cheapest 4 card suit."""
        suits = []
        for suit in self.suits:
            if self.suit_length(suit) == 4:
                suits.append(suit)
        cheapest = None
        for suit in suits:
            if cheapest:
                if suit < cheapest:
                    cheapest = suit
            else:
                cheapest = suit
        return cheapest

    def _biddable_five_card_major(self):
        """Return a biddable 5 card major."""
        suit = None
        if (self.hearts >= 5 and
                self.heart_suit not in self.opponents_suits):
            suit = self.heart_suit
        elif (self.spades >= 5 and
              self.spade_suit not in self.opponents_suits):
            suit = self.spade_suit
        return suit

    def _five_five_second_suit(self):
        """Return second suit in five/five hands."""
        if self.bid_one.is_suit_call:
            suit_one = self.bid_one.denomination
        else:
            suit_one = self.longest_suit
        if suit_one == self.longest_suit:
            suit_two = self.second_suit
        else:
            suit_two = self.second_suit
        if suit_two in self.opponents_suits:
            suit = suit_one
        elif suit_one in self.opponents_suits:
            suit = suit_two
        else:
            suit = suit_two
        return suit

    def _get_overcallers_level(self):
        """Return the level at which overcaller has bid."""
        if self.bid_one. is_value_call:
            suit = self.bid_one.denomination
            overcallers_level = self.next_level(suit)
        else:
            overcallers_level = 0
        return overcallers_level

    def _get_best_suit_for_rebid(self):
        """Return best own suit."""
        suit = self.longest_suit
        if self.five_five:
            suit = self.cheaper_suit(self.longest_suit, self.second_suit)
        return suit

    def _has_opponents_suit(self, opponents_bid):
        """Return True if hand has 4+ in opponents suit."""
        value = False
        if Bid(opponents_bid).is_suit_call:
            suit = Bid(opponents_bid).denomination
            if (self.suit_length(suit) >= 4 and
                    self.suit_points(suit) >= 4):
                value = True
            elif self.suit_length(suit) >= 5:
                value = True
        return value

    def _select_suit_for_rebid(self):
        """Select best suit for rebid."""
        suit_for_rebid = self.longest_suit
        if (self.advancer_bid_one.denomination.is_major and
                self.suit_length(self.advancer_bid_one.denomination) >= 4):
            suit = self.advancer_bid_one.denomination
        elif (self.opener_bid_one.denomination == suit_for_rebid or
              self.responder_bid_one.denomination == suit_for_rebid):
            suit = self.second_suit
        else:
            suit = suit_for_rebid
        return suit

    def _get_level_with_support_for_advancer(self):
        """Return the level to bid if overcaller can support advancer."""
        level = self.next_level(self.advancer_bid_one.denomination)
        if self._is_strong_or_advancer_has_jumped():
            level += 1
        if self.advancer_bid_one.denomination.is_major:
            level = min(level, 4)
        return level

    # Various boolean functions

    def _can_double(self):
        """Return True if hand is suitable for DOUBLE."""
        value = False
        overcallers_level = self._get_overcallers_level()
        opponents_agreed_suit = self.openers_agreed_suit()
        if opponents_agreed_suit:
            if (opponents_agreed_suit.is_suit and
                    not self.bid_one.is_pass and
                    self.advancer_bid_one.is_pass and
                    self.shape in [[4, 4, 4, 1], [5, 4, 4, 0]] and
                    not self.responder_bid_one.is_game and
                    not self.opener_bid_two.is_game and
                    self.next_level(self.spade_suit) <= 4 and
                    (overcallers_level <= 3 and
                     opponents_agreed_suit != self.spade_suit)):
                value = True
        if self.opener_bid_two.is_nt:
            value = False
        return value

    def _very_strong_six_card_suit_advancer_bids_game(self):
        """Return True if very strong, six card suit and advancer bids over responder."""
        result = (self.my_last_bid.is_double and
                  self.responder_bid_one.is_value_call and
                  self.advancer_bid_one.is_game and
                  self.hcp >= 23 and
                  self.shape[0] >= 6 and
                  self.next_level(self.longest_suit) <= 6)
        return result

    def _advancer_supports_overcaller_suit(self):
        """Return True if advancer support overcaller."""
        result = (self.bid_one.denomination == self.advancer_bid_one.denomination and
                  self.bid_one.is_suit_call)
        return result

    def _has_long_suit(self):
        """Return True with long suit."""
        level = self.next_level(self.longest_suit)
        result = (self.shape[0] >= 7 or
                  (self.shape[0] >= 6 and level <= 2))
        return result

    def _can_rebid_longest_suit_at_level_two(self):
        """Return True if able to rebid longest suit."""
        result = (self.hcp >= 12 and
                  not self.bid_one.is_pass and
                  self.suit_length(self.longest_suit) >= 5 and
                  self.next_level(self.longest_suit) <= 2 and
                  self.suit_points(self.longest_suit) >= 6 and
                  self.longest_suit not in self.opponents_suits and
                  self.responder_bid_one.is_pass and
                  not self.opener_bid_two.is_nt and
                  self.opener_bid_one.denomination == self.opener_bid_two.denomination)
        return result

    def _can_bid_second_suit_at_level_two(self):
        """Return True if able to bid second suit."""
        result = (self.hcp >= 12 and
                  not self.bid_one.is_pass and
                  self.suit_length(self.longest_suit) >= 5 and
                  self.suit_length(self.second_suit) >= 4 and
                  self.second_suit.rank < self.longest_suit.rank and
                  self.next_level(self.second_suit) <= 2 and
                  self.suit_points(self.second_suit) >= 2 and
                  self.second_suit not in self.opponents_suits and
                  self.responder_bid_one.is_pass and
                  not self.opener_bid_two.is_nt and
                  self.opener_bid_one.denomination == self.opener_bid_two.denomination)
        return result

    def _can_rebid_second_suit_at_level_two(self):
        """Return True if second suit is suitable for a rebid."""
        result = (self.hcp >= 14 and
                  self.next_level(self.second_suit) <= 2 and
                  self.five_four and
                  self.second_suit not in self.opponents_suits)
        return result

    def _can_rebid_nt_at_three_level(self):
        """Return True if can rebid nt at 3 level."""
        result = (self.bid_one.name != '1NT' and
                  self.advancer_bid_one.name == '2NT' and
                  self.hcp >= 14 and
                  self.nt_level <= 3)
        return result

    def _is_competitive_auction(self):
        """Return True if competitive auction."""
        result = (self.opener_bid_two.is_value_call or
                  self.responder_bid_one.is_value_call or
                  self.opener_bid_one.level > 1)
        return result

    def _strong_with_biddable_five_card_major(self):
        """Return True if strong with biddable 5 card major."""
        result = (self.five_card_major_or_better and
                  self.hcp >= 15 and
                  self._biddable_five_card_major())
        return result

    def _strong_and_advancer_bids_two_nt(self):
        """Return True if strong and advancer bids 2NT."""
        result = (self.advancer_bid_one.name == '2NT' and
                  self.hcp >= 15 and
                  self.nt_level <= 3)
        return result

    def _can_show_second_suit_after_advancer_has_bid(self):
        """Return True if can bid second suit and advancer has bid."""
        result = (self.shape[1] == 4 and
                  self.suit_points(self.second_suit) >= 5 and
                  self.second_suit not in self.opponents_suits and
                  not self.advancer_bid_one.is_pass)
        return result

    def _can_bid_six_card_suit_in_competitive_auction(self):
        """Return True if can bid 6 card suit in competitive auction."""
        result = (self.last_bid.is_value_call and
                  self.shape[0] >= 6 and
                  self.suit_points(self.longest_suit) >= 9 and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _can_rebid_suit_after_advancer_has_bid(self):
        """Return True if advancer bids and can repeat suit."""
        result = (self.hcp + self.shape[0] >= 16 and
                  (self.next_level(self.bid_one.denomination) <= 3 or
                   (self.shape[0] >= 7 and
                    self.advancer_bid_one.is_value_call)))
        return result

    def _weak_hand_after_opener_bids_weak_two(self):
        """Return True if opener bid weak two and moderate hand."""
        result = (self.opener_bid_one.level > 1 and
                  ((self.hcp <= 15 and self.shape[0] <= 5) or
                   (self.hcp <= 14 and self.shape[0] <= 6)) and
                  not self.responder_bid_one.is_value_call)
        return result

    def _strong_and_no_support_for_advancer(self, suit_for_rebid):
        """Return True if strong and no support for advancer."""
        result = (self.hcp >= 16 and
                  self.shape[0] >= 5 and
                  suit_for_rebid.is_major and
                  self.suit_length(self.advancer_bid_one.denomination) <= 3 and
                  suit_for_rebid not in self.opponents_suits and
                  self.next_level(suit_for_rebid) <= 3)
        return result

    def _very_strong_and_can_bid_nt(self):
        """Return True if very strong and can bid NT."""
        result = (self.hcp >= 20 and
                  self.advancer_bid_one.denomination.is_minor and
                  self.is_semi_balanced and
                  self.stoppers_in_bid_suits and
                  self.next_level(self.no_trumps) <= 3)
        return result

    def _can_support_advancer_after_double(self):
        """Return True if can support advancer."""
        result = (self.suit_length(self.advancer_bid_one.denomination) >= 3 and
                  not (self.shape[0] >= 6 and
                       self.longest_suit != self.advancer_bid_one.denomination))
        return result

    def _suit_for_rebid_is_minor_after_double(self, suit_for_rebid):
        """Return True if rebid suit is minor and can support advancer."""
        result = (suit_for_rebid.is_minor and
                  self.advancer_bid_one.level >= 2 and
                  self.advancer_bid_one.denomination.is_major and
                  self.suit_length(self.advancer_bid_one.denomination) >= 3)
        return result

    def _strong_and_can_bid_rebid_suit(self, suit_for_rebid):
        """Return True if strong and can rebid suit."""
        result = (self.hcp >= 15 and self.shape[0] >= 5 and
                  self.suit_points(suit_for_rebid) >= 5 and
                  suit_for_rebid not in self.opponents_suits and
                  (self.opener_bid_one.level == 2 or
                   self.next_level(suit_for_rebid) <= 2))
        return result

    def _can_bid_five_card_major_at_three_level(self):
        """Return True if 5 card major at 3 level."""
        result = (self.longest_suit.is_major and
                  self.shape[0] >= 5 and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _strong_balanced_and_advancer_bids_minor(self):
        """Return True if strong, balanced and advancer bids minor."""
        result = (self.hcp >= 17 and
                  not (self.bid_one.is_double and self.overcaller_in_fourth_seat) and
                  self.advancer_bid_one.denomination.is_minor and
                  self.is_balanced and
                  self.stoppers_in_bid_suits and
                  self.nt_level <= 3)
        return result

    def _above_minimum_and_advancer_bids_after_double(self):
        """Return True if above minimum and advancer at level two."""
        result = (self._strong_or_competitive_after_double() and
                  (self._jump_after_double() or self.responder_bid_one.is_pass) and
                  (self.hcp >= 14 or self.opener_bid_two.is_value_call) and
                  self.suit_length(self.advancer_bid_one.denomination) >= 4)
        return result

    def _strong_or_competitive_after_double(self):
        """Return True if strong  or competitive."""
        competitive_auction = Bid(self.bid_history[-1]).is_value_call
        result = (self.hcp >= 13 and competitive_auction and
                  (self.advancer_bid_one.level >= 3 or
                   self.suit_length(self.advancer_bid_one.denomination) >= 4) or
                  self.hcp >= 16)
        return result

    def _jump_after_double(self):
        """Return True if advancer has jumped after a double."""
        last_bid = None
        for bid in self.bid_history:
            if Bid(bid).is_value_call:
                last_bid = Bid(bid)
            if bid == 'D':
                break
        jump_steps = self.levels_to_bid(last_bid, self.advancer_bid_one)
        if jump_steps > 1:
            return True
        else:
            return False

    def _is_balanced_with_stopper(self, suit_for_rebid):
        """Return True if _balanced_with_stoppers."""
        result = (self.stoppers_in_bid_suits and
                  self.shape[0] == 4 and
                  suit_for_rebid.is_minor and
                  self.nt_level <= 3)
        return result

    def _four_card_support_for_advancers_major(self):
        """Return True if 4 card support for advancer."""
        result = (self.advancer_bid_one.denomination.is_major and
                  self.suit_length(self.advancer_bid_one.denomination) >= 4 or
                  (self.responder_bid_one.is_value_call and
                   self.suit_length(self.advancer_bid_one.denomination) == 3))
        return result

    def _can_bid_five_card_major_at_two_level(self, suit_for_rebid):
        """Return True if can bid 5 card major at two level."""
        result = (self.is_balanced and
                  suit_for_rebid.is_major and
                  self.suit_length(self.longest_suit) >= 5 and
                  self.next_level(suit_for_rebid) <= 2)
        return result

    def _has_five_card_suit_and_no_support_from_advancer(self, suit_for_rebid):
        """Return True if 5 card suit and no support from advancer."""
        result = (self.advancer_bid_one.denomination != suit_for_rebid and
                  self.shape[0] >= 5)
        return result

    def _is_balanced_and_long_suit_is_minor(self, suit_for_rebid):
        """Return True if suit is minor and balanced."""
        result = (suit_for_rebid.is_minor and
                  self.is_balanced and
                  self.stoppers_in_bid_suits)
        return result

    def _has_three_card_support_for_advancers_major(self, suit_for_rebid):
        """Return True if 3 card support for advancers suit."""
        result = (suit_for_rebid.is_minor and
                  self.advancer_bid_one.denomination.is_major and
                  self.suit_length(self.advancer_bid_one.denomination) >= 3)
        return result

    def _strong_and_opposition_weak(self):
        """Return True if 16 points and opposition weak."""
        result = (self.hcp >= 16 and
                  self.next_level(self.longest_suit) <= 3 and
                  self.opener_bid_two.is_pass and
                  self.opener_bid_one.is_suit_call and
                  not self.responder_bid_one.is_nt and
                  not self.responder_bid_one.is_double and
                  not self.responder_bid_one.is_redouble and
                  not self.five_five)
        return result

    def _is_strong_has_five_five(self):
        """Return True if strong with 5/5."""
        result = (self.hcp >= 16 and
                  self.five_five and
                  self.next_level(self.longest_suit) <= 3 and
                  (self.longest_suit not in self.opponents_suits or
                   self.second_suit not in self.opponents_suits))
        return result

    def _is_strong_with_good_five_card_suit(self):
        """Return True if strong with good 5 card suit."""
        result = (self.hcp >= 15 and
                  self.shape[0] >= 5 and
                  self.suit_points(self.longest_suit) >= 3 and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits and
                  (self.advancer_bid_one.is_value_call or
                   self.next_level(self.longest_suit) <= 2 or
                   self.hcp >= 19))
        return result

    def _is_strong_and_balanced(self):
        """Return True if strong and balanced."""
        result = (self.hcp >= 18 and
                  self.is_balanced and
                  self.nt_level <= 2 and
                  self.responder_bid_one.level == 1 and
                  self.stoppers_in_bid_suits)
        return result

    def _is_medium_and_balanced(self):
        """Return True if medium balanced."""
        result = (self.hcp >= 16 and
                  self.is_balanced and
                  self.nt_level <= 2 and
                  not self.opener_bid_one.is_nt and
                  self.responder_bid_one.is_pass and
                  self.stoppers_in_bid_suits)
        return result

    def _advancer_bids_stayman(self):
        """Return True if advancer bids stayman."""
        result = (self.bid_one.name == '1NT' and
                  self.advancer_bid_one.denomination == self.club_suit and
                  not Bid(self.bid_history[-3]).is_value_call)
        return result

    def _has_long_strong_suit(self):
        """Return True if long strong suit."""
        result = (self.suit_points(self.longest_suit) >= 5 and
                  (self.shape[0] >= 6 and self.hcp >= 13) or
                  (self.shape[0] >= 7 and self.hcp >= 12))
        return result

    def _has_five_four_and_can_bid_second_suit(self):
        """Return True if 5/4 and can bid second suit."""
        result = (self.five_four and
                  self.advancer_bid_one.level >= 2 and
                  self.hcp >= 12 and
                  self.next_level(self.second_suit) <= 2 and
                  self.second_suit not in self.opponents_suits)
        return result

    def _has_six_card_suit_and_no_support_for_advancer(self):
        """Return True if six card suit and no support for advancer."""
        result = (self.shape[0] >= 6 and
                  self.suit_length(self.advancer_bid_one.denomination) <= 2 and
                  self.bid_one.denomination.is_major and
                  self.next_level(self.bid_one.denomination) <= 3)
        return result

    def _can_bid_five_five_at_level_two(self):
        """Return True if 5/5 at level 2."""
        result = (self.five_five and
                  self.advancer_bid_one.is_suit_call and
                  self.next_level(self.second_suit) <= 2 and
                  self.second_suit not in self.opponents_suits)
        return result

    def _can_bid_seven_card_suit_at_level_three(self):
        """Return True if can bid 7 card suit at level 3."""
        result = (self.shape[0] >= 7 and
                  self.advancer_bid_one.is_value_call and
                  self.next_level(self.bid_one.denomination) <= 3)
        return result

    def _can_bid_sevel_card_suit_at_level_three(self):
        """Return True if can bid a 7 card suit at level 3."""
        result = (self.shape[0] >= 7 and
                  self.hcp <= 9 and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_twelve_points_no_support_for_advancer(self):
        """Return True if 12 points and no support for advancer."""
        result = (self.advancer_bid_one.level >= 2 and
                  self.suit_length(self.advancer_bid_one.denomination) <= 1 and
                  self.hcp >= 12 and
                  self.nt_level <= 3)
        return result

    def _advancer_has_jumped_not_competitive(self):
        """Return True if advancer has jumped and not competitive."""
        result = (self.hcp >= 13 and
                  (self.is_jump(Bid(self.bid_history[-3]), self.advancer_bid_one) or
                   (self.is_jump(self.bid_one, self.advancer_bid_one) and
                    Bid(self.bid_history[-3]).is_pass)))
        return result

    def _advancer_has_bid_at_level_three(self):
        """Return True if advancer bids at level 3."""
        result = (self.advancer_bid_one.level == 3 and
                  self.hcp >= 14 and
                  self.longest_suit.is_major and
                  not self.partner_has_passed and
                  self.next_level(self.bid_one.denomination) <= 4)
        return result

    def _advancer_has_jumped_over_openers_rebid(self):
        """Return True if advancer has jumped over opener bid two."""
        result = (self.is_jump(self.opener_bid_two, self.advancer_bid_one) and
                  self.hcp >= 11 and
                  self.opener_bid_two.is_suit_call and
                  self.next_level(self.longest_suit) <= 4)
        return result

    def _has_thirteen_points_and_can_bid_five_card_suit_at_level_three(self):
        """Return True if 13 points and can bid 5 card suit at 3 level."""
        result = (self.hcp >= 13 and
                  self.shape[0] >= 5 and
                  self.next_level(self.longest_suit) <= 3 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _has_eleven_points_and_can_bid_seven_card_suit_at_level_three(self):
        """Return True if 11 points and can bid 7 card suit at 3 level."""
        result = (self.hcp >= 11 and
                  self.shape[0] >= 7 and
                  self.next_level(self.bid_one.denomination) <= 3)
        return result

    def _has_fifteen_points_can_bid_major(self):
        """Return True if strong and can bid major."""
        result = (self.support_points(self.bid_one.denomination) >= 15 and
                  self.bid_one.denomination.is_major and
                  self.bid_one.denomination not in self.opponents_suits)
        return result

    def _can_raise_after_advancer_jumps(self, level):
        """Return True if can bid again after advancer has jumped."""
        result = (not self.is_insufficient_bid(self.suit_bid(level, self.bid_one.denomination)) and
                  self.bid_one.denomination not in self.opponents_suits)
        return result

    def _can_attempt_support(self, level):
        """Return True if can attempt to support advancer."""
        result = (level <= self.suit_length(self.advancer_bid_one.denomination) or
                  self.hcp >= 19 or
                  self.advancer_bid_one.level == 3 or
                  self.is_jump(self.my_last_bid, self.advancer_bid_one))
        return result

    def _advancer_has_bid_clubs_and_can_support(self, level):
        """Return True if advancer has bid clubs."""
        result = (level >= self.next_level(self.advancer_bid_one.denomination) and
                  not (self.my_last_bid.is_nt and
                  self.advancer_bid_one.denomination == self.club_suit) and
                  self.support_points(self.advancer_bid_one.denomination) >= 13 and
                  self._can_attempt_support(level))
        return result

    def _is_strong_or_advancer_has_jumped(self):
        """Return True if strong or advancer has jumped."""
        result = (self.hcp >= 17 or
                  (self.suit_length(self.advancer_bid_one.denomination) >= 4 and
                   self.support_shape_points(self.advancer_bid_one.denomination) >= 3) or
                  (self.is_jump(self.my_last_bid, self.advancer_bid_one) and
                   self.advancer_bid_one.denomination.is_major))
        return result

    def _can_show_four_card_suit_at_level_two(self):
        """Return True if can show four card suit at level 4."""
        result = (self.opener_bid_one.is_suit_call and
                  self.shape[0] >= 6 and
                  self.shape[1] >= 4 and
                  self.next_level(self.second_suit) <= 2 and
                  self.longest_suit not in self.opponents_suits and
                  self.second_suit not in self.opponents_suits)
        return result

    def _can_show_four_card_suit_at_level_four(self):
        """Return True if can show four card suit at level 4."""
        result = (self.opener_bid_one.is_suit_call and
                  self.shape[1] >= 4 and
                  3 <= self.next_level(self.second_suit) <= 4 and
                  self.longest_suit not in self.opponents_suits and
                  self.second_suit not in self.opponents_suits)
        return result

    def _is_strong_in_opponents_suit(self):
        """Return True if strong in opponents suit."""
        result = (self.suit_length(self.last_bid.denomination) >= 4 and
                  self.suit_points(self.last_bid.denomination) >= 4)
        return result

    def _is_weak_and_some_support_for_advancer(self, suit_for_rebid):
        """Return True if weak and some support for partner."""
        result = (self.hcp <= 16 and
                  self.advancer_bid_one.level >= 2 and
                  self.suit_length(self.advancer_bid_one.denomination) >= 2 and
                  self.last_bid.is_pass and
                  self.next_level(suit_for_rebid) <= 2)
        return result

    def _advancer_bids_three_nt_and_long_suit(self):
        """Return True if advancer has bid 3NT and long suit."""
        suit = self.longest_suit
        result = (self.advancer_bid_one.name == '3NT' and
                  self.shape[0] >= 7 and
                  self.game_level(suit) >= self.next_level(suit) and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _advancer_bids_two_nt_and_six_card_suit_and_weak(self):
        """Return True if advancer has bid 2NT and single suited and weak."""
        result = (self.advancer_bid_one.name == '2NT' and
                  self.shape[0] >= 6 and
                  self.hcp <= 9 and
                  self.longest_suit not in self.opponents_suits)
        return result

    def _partner_weak_after_double(self):
        """Return True if partner weak and balanced."""
        result = ((self.partner_bids_at_lowest_level() or self.hcp < 19) and
                  Bid(self.bid_history[-1]).is_pass and
                  not self.four_four_four_one)
        return result

    def _can_raise_advancer_after_double(self):
        """Return True if can raise advancer after a double."""
        advancer_not_weak = ((self.my_last_bid.level == 2 and
                             not self.partner_bids_at_lowest_level()) or
                             self.opener_bid_one.is_nt)
        result = (self.hcp >= 16 and
                  advancer_not_weak and
                  self.right_hand_bid.is_pass and
                  not self.next_level_bid(self.advancer_bid_one.denomination).is_game)
        return result

    def _can_bid_second_suit(self):
        """Return True if can bid second suit."""
        result = (self.shape[1] >= 5 and
                  self.next_level(self.second_suit) <= 2 and
                  self.second_suit not in self.opponents_suits)
        return result

    def _strong_over_weak_three(self):
        """Return True if strong over weak 3."""
        result = (self.hcp >= 20 and self.opponents_at_game and
                  self.opener_bid_one.level == 3 and
                  self.suit_points(self.opener_bid_one.denomination) >= 4)
        return result

    def _support_for_partners_level_three_call(self):
        """Return True if Support for partners level 3 bid."""
        result = (self.partner_bid_one.level == 3 and
                  self.suit_length(self.partner_bid_one.denomination) >= 3 and
                  self.bid_history[-3] != 'P')
        return result

    def _very_strong_and_some_support_for_partner(self):
        """Return True if very strong and some support for partner."""
        result = (self.hcp >= 23 and
                  ((self.partner_last_bid.denomination.is_major and
                    self.suit_length(self.partner_last_bid.denomination) >= 2) or
                  ((self.partner_last_bid.denomination.is_minor and
                    self.suit_length(self.partner_last_bid.denomination) >= 4))))
        return result

    def _minimum_after_double_over_1nt(self):
        """Return True if minimum after double over 1NT."""
        result = (self.opener_bid_one.name == '1NT' and
                  self.hcp == 16 and
                  self.responder_bid_one.is_value_call and
                  self.advancer_bid_one.level >= 3)
        return result

    def _good_second_suit_and_rebiddable_long_suit(self):
        """Return True if good second suit and rebiddable first suit."""
        result = (self.shape[1] >= 4 and
                  self.suit_points(self.longest_suit) >= 4 and
                  self.suit_points(self.second_suit) >= 5 and
                  self.second_suit.rank < self.longest_suit.rank and
                  self.second_suit not in self.opponents_suits)
        return result
